#! /usr/bin/env python

import numpy as np
import argparse
import os
from astropy.io import fits
import subprocess as sp
import scipy
import sys
import copy

from scipy.special import gamma, gammainc, gammaincinv

from scipy.optimize import bisect, fmin, newton

import matplotlib.pyplot as plt

from scipy.interpolate import UnivariateSpline

from galfitools.galin.galfit import Galfit, conver2Sersic, SelectGal, numComps, GetRadAng


from galfitools.galin.galfit import GalComps, GalHead



def getBreak(galfitFile: str, dis: int, inicomp: int, quick: bool, random: int, num_comp: int, angle: float, plot: bool, ranx: list)-> float:
    '''gets the break radius from a set of Sersics'''


   

    #head = ReadHead(galfitFile)
    #galcomps = ReadComps(galfitFile)


    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()



    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 



    comps.Flux = 10**((-comps.Mag)/2.5)

    k = gammaincinv(2*comps.Exp, 0.5)

    denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
    denom2 = (comps.Exp)*(k**(-2*comps.Exp))
    denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 

    denom = denom1*denom2*denom3 
    
    comps.Ie = comps.Flux/denom



    comps = SelectGal(comps, dis, num_comp)

    #taking the last component position angle for the whole galaxy

    maskgal = (comps.Active == True) 

    if angle:
        theta = angle
    else:
        theta = comps.PosAng[maskgal][-1]  




    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute Re')
        print('exiting..')
        sys.exit(1)



    #########################
    ### computing the slope
    #########################


    if plot:

        if ranx:
            (xmin,xmax)=ranx[0], ranx[1]
        else:
            xmin = 0.1
            xmax = 100

        R = np.arange(xmin,xmax,.1)

        gam = GetBreak().GalBreak(R, comps, theta) 


        plt.plot(R, gam)

        plt.xlabel("Rad")
        plt.ylabel("Second derivative")

        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("Break.png")


    if quick:

        rbreak = GetBreak().FindBreak(comps, theta, inicomp) 

        line = 'The break radius  is {:.2f} pixels \n'.format(rbreak)
        #print(line)

    else:

        if random:

            radius = np.random.random(random)*comps.Rad[maskgal][inicomp] 
            #print('The initial search radius are: ',radius)

        else:
    
            radius = comps.Rad[maskgal]
            #print('The initial search radius are the effective radius of the components')


        rbreak = MulFindBreak(comps, theta, radius) 



    return rbreak, N, theta


def MulFindBreak(comps, theta, radius):

    maskgal = (comps.Active == True) 

    radsbreak = GetBreak().MulFindBreak(comps, theta, radius) 

    betas = np.array([])

    #print('finding global minimum')

    for r in radsbreak:

        beta = GetBreak().BreakSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)

        
        betas = np.append(betas, beta)


    radmask = (radsbreak < radius[-1]) #hope it works

    idx = np.where(betas == min(betas))  

    
    return radsbreak[idx][0]


class GetBreak:
    '''class to obtain the break radius for the whole galaxy'''



    def FullSlopeSer(self, R: float, Re: list, n: list, q: list, pa: list, theta: float) -> float:


        SlptotR = self.SlopeSer(R, Re, n, q, pa, theta) 

        return SlptotR.sum()


    def GalBreak(self, R: list, comps: GalComps, theta: float) -> float:
        "plots the curvature of the function"

        maskgal = (comps.Active == True) #using active components only 

        kappa = np.array([])

        for r in R:

            beta = self.BreakSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)

            kappa = np.append(kappa, beta)


        return kappa 


    def funGalSlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float, slope: float) -> float:


        fun = self.SlopeSer(R, Ie, Re, n, q, pa, theta)  - slope


        return fun
     

    def FindBreak(self, comps: GalComps, theta: float, initial_comp: int) -> float:
        "return the break radii of a set of Sersic functions"

        maskgal = (comps.Active == True) #using active components only 

        init = comps.Rad[maskgal][initial_comp] #component as initial value, hope it works 

        breakrad = scipy.optimize.fmin(self.funGalBreakSer, init, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta))


        return breakrad[0] 


    def MulFindBreak(self, comps: GalComps, theta: float, radius: list) -> float:
        "return the break radius evaluated at different effective radius"


        brads = np.array([])

        maskgal = (comps.Active == True) #using active components only 


        for idx, item in enumerate(radius):

            init = item 

            breakrad = scipy.optimize.fmin(self.funGalBreakSer, init, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta))


            line = 'Optimized radius: {:.2f} \n '.format(breakrad[0])
            print(line)


            brads = np.append(brads, breakrad[0])


        return brads 
 



    def funGalBreakSer(self, R, Ie, Re, n, q, pa, theta):


        beta = self.BreakSer(R, Ie, Re, n, q, pa, theta)

           

        return beta 


    def FindSlope(self, comps: GalComps, theta: float, slope: float) -> float:
        "return the Re of a set of Sersic functions. It uses Bisection"

        maskgal = (comps.Active == True) #using active components only 

        a = 0.1 
        b = comps.Rad[maskgal][-1] * 10  # hope it doesn't crash

        try:
            Radslp = bisect(self.funGalSlopeSer, a, b, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta, slope))

        except:
            print("unable to solve equation in the given range. Setting n to 99")
            Radslp = 0 

        return Radslp 


    def var_X(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varx = k*((R/Re)**(1/n) - 1)


        return varx

    def var_Xprim(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr = (k/n)*((R/Re)**(1/n - 1))*(1/Re) 

        return varxpr

    def var_Xprim2(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr2 = (k/n)*((R/Re)**(1/n - 2))*(1/Re**2)*(1/n -1) 

        return varxpr2


    def var_S(self, R: float, Ie: list,  Re: list, n: list, X: list):

        S = Ie*np.exp(-X)

        return S.sum()

    def var_Sprim(self, R: float, Ie: list,  Re: list, n: list, X: list, Xprim: list):

        Sprim = -Ie*np.exp(-X)*Xprim

        return Sprim.sum()
    
    def var_Sprim2(self, R: float, Ie: list, Re: list, n: list, X: list, Xprim: list, Xprim2: list):

        Sprim2 = Ie*np.exp(-X)*Xprim**2 - Ie*np.exp(-X)*Xprim2 

        return Sprim2.sum()


    def BreakSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''Break from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        Xprim = self.var_Xprim(Rcor, Re, n)
        Xprim2 = self.var_Xprim2(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        Sprim = self.var_Sprim(Rcor, Ie, Re, n, X, Xprim)
        Sprim2 = self.var_Sprim2(Rcor, Ie, Re, n, X, Xprim, Xprim2)

        Break = (Sprim/S)*(R/np.log10(np.e)) + (
                    (Sprim2*S - Sprim**2)/S**2)*(R**2/np.log10(np.e)) 

        return Break 

    def SlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''slope from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        Xprim = self.var_Xprim(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        Sprim = self.var_Sprim(Rcor, Ie,  Re, n, X, Xprim)

        Slp = (Sprim/S)*R 


        return Slp




def getBreak2(galfitFile: str, dis: int, angle: float, num_comp: int, plot: bool, ranx: list) -> float:
    '''gets the break radius from a set of Sersics using another method'''


    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()



    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 



    comps.Flux = 10**((-comps.Mag)/2.5)

    k = gammaincinv(2*comps.Exp, 0.5)

    denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
    denom2 = (comps.Exp)*(k**(-2*comps.Exp))
    denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 

    denom = denom1*denom2*denom3 
    
    comps.Ie = comps.Flux/denom



    comps = SelectGal(comps, dis, num_comp)


    maskgal = (comps.Active == True) 
    if angle:
        theta = angle
    else:
        #taking the last component position angle for the whole galaxy
        theta = comps.PosAng[maskgal][-1]  


    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute Re')
        print('exiting..')
        sys.exit(1)



    #########################
    ### computing the slope
    #########################
    if ranx:
        (xmin,xmax)=ranx[0], ranx[1]
    else:
        xmin = 0.1
        xmax = 100


    R = np.arange(xmin,xmax,.1)

    lrad= np.log10(R)
    slp = GetSlope().GalSlope(R, comps, theta) 


    if plot:
        plt.close()
        plt.plot(R, slp)

        plt.xlabel("Rad")
        plt.ylabel("Slope")

        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("FirstDerivative.png")


    ###
    yspl = UnivariateSpline(lrad,slp,s=0,k=4)

    yspl2d = yspl.derivative(n=2)
    yspl1d = yspl.derivative(n=1)

    if plot:
        plt.close()
        plt.plot(lrad,yspl1d(lrad))
        plt.xlabel("Log(Rad)")
        plt.ylabel("Second derivative")

        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("SecondDerivative.png")




    idx = np.where(yspl1d(lrad) == max(yspl1d(lrad)))[0][0]

    brad = lrad[idx]




    rbreak = 10**brad



    return rbreak, N, theta  
############################
############################
############################



def getFWHM(galfitFile: str, dis: int, angle: float, num_comp: int):
    '''gets the FWHM from a set of Sersics'''


    #head = ReadHead(galfitFile)
    #galcomps = ReadComps(galfitFile)


    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()
 


    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 



    comps.Flux = 10**((-comps.Mag)/2.5)

    k = gammaincinv(2*comps.Exp, 0.5)

    denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
    denom2 = (comps.Exp)*(k**(-2*comps.Exp))
    denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 

    denom = denom1*denom2*denom3 
    
    comps.Ie = comps.Flux/denom



    comps = SelectGal(comps, dis, num_comp)

    #taking the last component position angle for the whole galaxy

    maskgal = (comps.Active == True) 

    if angle:
        theta = angle
    else:
        theta = comps.PosAng[maskgal][-1]  




    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute FWHM')
        print('exiting..')
        sys.exit(1)



    #########################
    ### computing the slope
    #########################



    fwhm = GetFWHM().FindFWHM(comps, theta) 


    return fwhm, N, theta



class GetFWHM:
    '''class to obtain the FWHM for the whole galaxy'''



    def FullFWHMSer(self, R: float, Re: list, n: list, q: list, pa: list, theta: float) -> float:


        SlptotR = self.FWHMSer(R, Re, n, q, pa, theta) 

        return SlptotR.sum()


    def GalFWHM(self, R: list, comps: GalComps, theta: float) -> float:

        maskgal = (comps.Active == True) #using active components only 

        gam = np.array([])

        for r in R:

            slp = self.FWHM(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)
            gam = np.append(gam, slp)



        return gam


    def funGalFWHMSer(self, R: float, S0: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:


        fun = self.FWHMSer(R, Ie, Re, n, q, pa, theta)  -  S0/2  


        return fun
     



    def FindFWHM(self, comps: GalComps, theta: float) -> float:
        "return the fwhm of a set of Sersic functions. It uses Bisection"

        maskgal = (comps.Active == True) #using active components only 

        #find the surface brightness at S(R=0)
        X = self.var_X(0, comps.Rad[maskgal], comps.Exp[maskgal])
        S0 = self.var_S(0, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], X)
 



        a = 0.001 
        b = comps.Rad[maskgal][-1] * 10  # hope it doesn't crash

        try:
            Radslp = bisect(self.funGalFWHMSer, a, b, args=(S0,comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta ))
        
        except:
            print("solution not found in given range")
            Radslp = 0


        return 2*Radslp 


    def var_X(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varx = k*((R/Re)**(1/n) - 1)

        return varx

    def var_Xprim(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr = (k/n)*((R/Re)**(1/n - 1))*(1/Re) 

        return varxpr

    def var_S(self, R: float, Ie: list,  Re: list, n: list, X: list):

        S = Ie*np.exp(-X)

        return S.sum()

    def var_Sprim(self, R: float, Ie: list,  Re: list, n: list, X: list, Xprim: list):

        Sprim = Ie*np.exp(-X)*Xprim

        return Sprim.sum()

    def FWHMSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''I(R) from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        #Xprim = self.var_Xprim(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        #Sprim = self.var_Sprim(Rcor, Ie,  Re, n, X, Xprim)

        Ifwhm = S 


        return Ifwhm 

############################
############################
############################



def getKappa(galfitFile: str, dis: int, inicomp: int, quick: bool, random: int, angle: float, num_comp: int, plot: bool, ranx: list) -> float:
    '''gets the Kappa radius from a set of Sersics'''


   

    #head = ReadHead(galfitFile)
    #galcomps = ReadComps(galfitFile)


    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()
 




    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 



    comps.Flux = 10**((-comps.Mag)/2.5)

    k = gammaincinv(2*comps.Exp, 0.5)

    denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
    denom2 = (comps.Exp)*(k**(-2*comps.Exp))
    denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 

    denom = denom1*denom2*denom3 
    
    comps.Ie = comps.Flux/denom



    comps = SelectGal(comps, dis, num_comp)

    #taking the last component position angle for the whole galaxy

    maskgal = (comps.Active == True) 
    if angle:
        theta = angle
    else:
        theta = comps.PosAng[maskgal][-1]  




    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute Re')
        print('exiting..')
        sys.exit(1)

    line = 'Using a theta value of : {:.2f} degrees\n'.format(theta)
    print(line)



    #########################
    ### computing the slope
    #########################


    if plot:

        if ranx:
            (xmin,xmax)=ranx[0], ranx[1]
        else:
            xmin = 0.1
            xmax = 100

        R = np.arange(xmin,xmax,.1)


        gam = GetKappa().GalKappa(R, comps, theta) 


        plt.plot(R, gam)
        plt.xlabel("Radius")
        plt.ylabel("Curvature")
        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("Kappa.png")


    if quick:

        rkappa = GetKappa().FindKappa(comps, theta, inicomp) 

        line = 'The Kappa radius  is {:.2f} pixels \n'.format(rbreak)
        #print(line)

    else:

        if random:

            radius = np.random.random(random)*comps.Rad[maskgal][inicomp] 
            #print('The initial search radius are: ',radius)

        else:
    
            radius = comps.Rad[maskgal]
            #print('The initial search radius are the effective radius of the components')


        rkappa = MultiFindKappa(comps, theta, radius) 



    return rkappa, N, theta 



def MultiFindKappa(comps, theta, radius):


    maskgal = (comps.Active == True) 

    radskappa = GetKappa().MulFindKappa(comps, theta, radius) 

    kappas = np.array([])

    print('finding global minimum:')

    for r in radskappa:

        kappa = GetKappa().funGalKappaSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)

        
        kappas = np.append(kappas, kappa)


    radmask = (radskappa < radius[-1]) #hope it works

    idx = np.where(kappas == min(kappas))  

    
    return radskappa[idx][0]




class GetKappa:
    '''class to obtain the Kappa radius for the whole galaxy'''



    def FullSlopeSer(self, R: float, Re: list, n: list, q: list, pa: list, theta: float) -> float:


        SlptotR = self.SlopeSer(R, Re, n, q, pa, theta) 

        return SlptotR.sum()


    def GalKappa(self, R: list, comps: GalComps, theta: float) -> float:
        "plots the curvature of the function"

        maskgal = (comps.Active == True) #using active components only 

        kappa = np.array([])

        for r in R:

            beta = self.BetaSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)

            gam = self.SlopeSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)

            kap = np.abs(beta)/(1 + gam**2)**(3/2)

            kappa = np.append(kappa, kap)


        return kappa 


    def funGalSlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float, slope: float) -> float:


        fun = self.SlopeSer(R, Ie, Re, n, q, pa, theta)  - slope


        return fun
     

    def FindKappa(self, comps: GalComps, theta: float, initial_comp: int) -> float:
        "return the break radius of a set of Sersic functions"

        maskgal = (comps.Active == True) #using active components only 

        init = comps.Rad[maskgal][initial_comp] #component as initial value, hope it works 

        kapparad = scipy.optimize.fmin(self.funGalKappaSer, init, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta))


        return kapparad[0] 
  

    def funGalKappaSer(self, R, Ie, Re, n, q, pa, theta):



        beta = self.BetaSer(R, Ie, Re, n, q, pa, theta)

        gam = self.SlopeSer(R, Ie, Re, n, q, pa, theta)
           
        kappa = np.abs(beta)/(1 + gam**2)**(3/2)


        return -kappa 


    def MulFindKappa(self, comps: GalComps, theta: float, radius: list) -> float:
        "return the kappa radius evaluated at different effective radius"


        krads = np.array([])

        maskgal = (comps.Active == True) #using only active components 


        for idx, item in enumerate(radius):

            init = item 

            kapparad = scipy.optimize.fmin(self.funGalKappaSer, init, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta))

            
            line = 'Optimized radius: {:.2f} \n '.format(kapparad[0])
            print(line)

            krads = np.append(krads, kapparad[0])


        return krads 
 






    def FindSlope(self, comps: GalComps, theta: float, slope: float) -> float:
        "return the Re of a set of Sersic functions. It uses Bisection"

        maskgal = (comps.Active == True) #using active components only 

        a = 0.1 
        b = comps.Rad[maskgal][-1] * 10  # hope it doesn't crash


        try:
            Radslp = bisect(self.funGalSlopeSer, a, b, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta, slope))

        except:
            print("solution not found in given range")
            Radslp = 0


        return Radslp 


    def var_X(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varx = k*((R/Re)**(1/n) - 1)


        return varx

    def var_Xprim(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr = (k/n)*((R/Re)**(1/n - 1))*(1/Re) 

        return varxpr

    def var_Xprim2(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr2 = (k/n)*((R/Re)**(1/n - 2))*(1/Re**2)*(1/n -1) 

        return varxpr2


    def var_S(self, R: float, Ie: list,  Re: list, n: list, X: list):

        S = Ie*np.exp(-X)

        return S.sum()

    def var_Sprim(self, R: float, Ie: list,  Re: list, n: list, X: list, Xprim: list):

        Sprim = -Ie*np.exp(-X)*Xprim

        return Sprim.sum()
    
    def var_Sprim2(self, R: float, Ie: list, Re: list, n: list, X: list, Xprim: list, Xprim2: list):

        Sprim2 = Ie*np.exp(-X)*Xprim**2 - Ie*np.exp(-X)*Xprim2 

        return Sprim2.sum()


    def BetaSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''Kappa from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        Xprim = self.var_Xprim(Rcor, Re, n)
        Xprim2 = self.var_Xprim2(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        Sprim = self.var_Sprim(Rcor, Ie, Re, n, X, Xprim)
        Sprim2 = self.var_Sprim2(Rcor, Ie, Re, n, X, Xprim, Xprim2)

        Beta = (Sprim/S)*(R/np.log10(np.e)) + (
                    (Sprim2*S - Sprim**2)/S**2)*(R**2/np.log10(np.e)) 

        return Beta 

    def SlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''slope from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        Xprim = self.var_Xprim(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        Sprim = self.var_Sprim(Rcor, Ie,  Re, n, X, Xprim)

        Slp = (Sprim/S)*R 


        return Slp





############################
############################
############################



def getReComp(galfitFile: str, dis: int, eff: float, angle: float, num_comp: int)-> float:
    '''gets the effective radius from a set of Sersics'''


    assert (eff > 0) and (eff <= 1), 'effrad must be a value between 0 and 1'


    #head = ReadHead(galfitFile)
    #galcomps = ReadComps(galfitFile)

    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()
 



    galcomps = SelectGal(galcomps, dis, num_comp)

    #taking the last component position angle for the whole galaxy

    maskgal = (galcomps.Active == True) 
    if angle:
        theta = angle
    else:
        theta = galcomps.PosAng[maskgal][-1]  


    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 


    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute Re')
        print('exiting..')
        sys.exit(1)



    EffRad, totmag = GetReff().GetReSer(head, comps, eff, theta)

    meanme = GetMe().MeanMe(totmag, EffRad*head.scale)
    me = GetMe().Me(head, comps, EffRad*head.scale, theta)

    #EffRad_arc = EffRad*head.scale

    return EffRad, totmag, meanme, me, N, theta 




class GetMe:
    '''Class to obtain the surface brightness at effective radius'''

    def MeanMe(self, magtot: float, effrad: float) -> float:

        meanme = magtot + 2.5*np.log10(2*np.pi*effrad**2)

        return meanme


    def Me(self, head, comps, EffRad, theta):

        comps.Rad = comps.Rad*head.scale
        comps.Flux = 10**((-comps.Mag)/2.5)

        k = gammaincinv(2*comps.Exp, 0.5)

        denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
        denom2 = (comps.Exp)*(k**(-2*comps.Exp))
        #denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 
        denom3 = (gamma(2*comps.Exp))

        denom = denom1*denom2*denom3 
        
        comps.Ie = comps.Flux/denom


        maskgal = (comps.Active == True) 
        Itotr = self.Itotser(EffRad, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta) 

        me = -2.5*np.log10(Itotr)

        return me

     
    def Itotser(self, R: float, Ie: list, rad: list, n: list, q: list, pa: list, theta: float) -> float:

        ItotR = self.Iser(R, Ie, rad, n, q, pa, theta) 

        return ItotR.sum()



    def Iser(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''sersic flux to a determined R'''


        k = gammaincinv(2*n, 0.5)

        Rcor = GetRadAng(R, q, pa, theta) 

        Ir = Ie*np.exp(-k*((Rcor/Re)**(1/n) - 1))

        
        return Ir



### Sersic components 
class GetReff:
    '''class to obtain the effective radius for the whole galaxy'''


    def GetRfracSer(self, head, comps, F, theta):



        rads = np.array([])

        for f in F:


            r, totmag = GetReff().GetReSer(head, comps, f, theta)

            rads = np.append(rads, r)



        return rads 





    def GetReSer(self, galhead: GalHead, comps: GalComps, eff: float, theta: float) -> float:


        maskgal = (comps.Active == True) 

        comps.Flux = 10**((galhead.mgzpt - comps.Mag)/2.5)

        totFlux = comps.Flux[maskgal].sum()

        totmag = -2.5*np.log10(totFlux) + galhead.mgzpt

        a = 0.1
        b = comps.Rad[maskgal][-1] * 1000  # hope it doesn't crash

        Reff = self.solveSerRe(a, b, comps.Flux[maskgal], comps.Rad[maskgal], 
                comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], totFlux, eff, theta)


        return Reff, totmag



    def solveSerRe(self, a: float, b: float, flux: list, rad: list, n: list, q: list, pa: list, totFlux: float, eff: float, theta: float) -> float:
        "return the Re of a set of Sersic functions. It uses Bisection"

        try:
            Re = bisect(self.funReSer, a, b, args=(flux, rad, n, q, pa, totFlux, eff, theta))

        except:
            print("solution not found in given range")
            Re = 0



        return Re


    def funReSer(self, R: float, flux: list, rad: list, n: list, q: list, pa: list, totFlux: float, eff: float, theta: float) -> float:
        

        fun = self.Ftotser(R, flux, rad, n, q, pa, theta) - totFlux*eff

        return fun
     
    def Ftotser(self, R: float, flux: list, rad: list, n: list, q: list, pa: list, theta: float) -> float:

        ftotR = self.Fser(R, flux, rad, n, q, pa, theta) 

        return ftotR.sum()



    def Fser(self, R: float, Flux: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''sersic flux to a determined R'''
        
        k = gammaincinv(2*n, 0.5)

        Rcor = GetRadAng(R, q, pa, theta) 

        X = k*(Rcor/Re)**(1/n) 

        Fr = Flux*gammainc(2*n, X) 
        
        return Fr


############################
############################
############################



def getSlope(galfitFile: str, dis: int, slope: float, angle: float, num_comp: int, plot: bool, ranx: list)-> float:
    '''gets the slope radius from a set of Sersics'''


    #head = ReadHead(galfitFile)
    #galcomps = ReadComps(galfitFile)

    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()
 
    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps = conver2Sersic(galcomps) 



    comps.Flux = 10**((-comps.Mag)/2.5)

    k = gammaincinv(2*comps.Exp, 0.5)

    denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
    denom2 = (comps.Exp)*(k**(-2*comps.Exp))
    denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 

    denom = denom1*denom2*denom3 
    
    comps.Ie = comps.Flux/denom



    comps = SelectGal(comps, dis, num_comp)

    #taking the last component position angle for the whole galaxy

    maskgal = (comps.Active == True) 

    if angle:
        theta = angle
    else:
        theta = comps.PosAng[maskgal][-1]  




    N = numComps(comps,'all')

    if N == 0:
        print('not enough number of components to compute Re')
        print('exiting..')
        sys.exit(1)



    #########################
    ### computing the slope
    #########################

    if plot:

        if ranx:
            (xmin,xmax)=ranx[0], ranx[1]
        else:
            xmin = 0.1
            xmax = 100

        R = np.arange(xmin,xmax,.1)


        gam = GetSlope().GalSlope(R, comps, theta) 


        plt.plot(R, gam)
        plt.xlabel("Radius")
        plt.ylabel("Slope")
        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("slope.png")


    rgam = GetSlope().FindSlope(comps, theta, slope) 


    return rgam, N, theta 




class GetSlope:
    '''class to obtain the effective radius for the whole galaxy'''



    def FullSlopeSer(self, R: float, Re: list, n: list, q: list, pa: list, theta: float) -> float:


        SlptotR = self.SlopeSer(R, Re, n, q, pa, theta) 

        return SlptotR.sum()


    def GalSlope(self, R: list, comps: GalComps, theta: float) -> float:

        maskgal = (comps.Active == True) #using active components only 

        gam = np.array([])

        for r in R:

            slp = self.SlopeSer(r, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta)
            gam = np.append(gam, slp)



        return gam


    def funGalSlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float, slope: float) -> float:


        fun = self.SlopeSer(R, Ie, Re, n, q, pa, theta)  - slope


        return fun
     



    def FindSlope(self, comps: GalComps, theta: float, slope: float) -> float:
        "return the Re of a set of Sersic functions. It uses Bisection"

        maskgal = (comps.Active == True) #using active components only 

        a = 0.1 
        b = comps.Rad[maskgal][-1] * 10  # hope it doesn't crash

        try:
            Radslp = bisect(self.funGalSlopeSer, a, b, args=(comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta, slope))

        except:
            print("solution not found in given range")
            Radslp = 0



        return Radslp 


    def var_X(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varx = k*((R/Re)**(1/n) - 1)


        return varx

    def var_Xprim(self, R: float, Re: list, n: list):

        k = gammaincinv(2*n, 0.5)

        varxpr = (k/n)*((R/Re)**(1/n - 1))*(1/Re) 


        return varxpr

    def var_S(self, R: float, Ie: list,  Re: list, n: list, X: list):

        S = Ie*np.exp(-X)

        return S.sum()

    def var_Sprim(self, R: float, Ie: list,  Re: list, n: list, X: list, Xprim: list):

        Sprim = Ie*np.exp(-X)*Xprim

        return Sprim.sum()

    def SlopeSer(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''slope from sersic function to a determined R'''
            
        Rcor = GetRadAng(R, q, pa, theta) 

        X = self.var_X(Rcor, Re, n)
        Xprim = self.var_Xprim(Rcor, Re, n)

        S = self.var_S(Rcor, Ie,  Re, n, X)
        Sprim = self.var_Sprim(Rcor, Ie,  Re, n, X, Xprim)

        Slp = (Sprim/S)*R 


        return Slp



############################
############################
############################



def getBulgeRad(galfitFile1, galfitFile2, dis, num_comp, angle, plot, ranx):
    '''gets the bulge radius or the radius where two models of surface brightness models are equal'''



    galfit1 = Galfit(galfitFile1)
    head1 = galfit1.ReadHead()
    galcomps1 = galfit1.ReadComps()


    galfit2 = Galfit(galfitFile2)
    head2 = galfit2.ReadHead()
    galcomps2 = galfit2.ReadComps()



    galcomps1 = SelectGal(galcomps1, dis, num_comp)

    galcomps2 = SelectGal(galcomps2, dis, num_comp)



    #taking the last component position angle for the whole galaxy

    maskgal1 = (galcomps1.Active == True) 
    maskgal2 = (galcomps2.Active == True) 

    if angle:
        theta = angle
    else:
        theta = galcomps2.PosAng[maskgal2][-1]  

    #theta = 18.2534 


    #convert all exp, gaussian and de vaucouleurs to Sersic format
    comps1 = conver2Sersic(galcomps1) 
    comps2 = conver2Sersic(galcomps2) 


    N1 = numComps(comps1,'all')
    N2 = numComps(comps2,'all')


    if (N1 == 0) or (N2 == 0):
        print('not enough number of components of one of the two models to proceed')
        print('exiting..')
        sys.exit(1)




    if plot:

        if ranx:
            (xmin, xmax) = ranx[0], ranx[1]
        else:
            xmin = 0.1
            xmax = 100



        R = np.arange(xmin,xmax,.1)


        Idiff = getDiff(head1, comps1, comps2, R, theta)

        plt.clf()
        plt.plot(R, Idiff)
        plt.grid(True)
        plt.minorticks_on()
        plt.xlabel("Rad")
        plt.ylabel("I1 - I2")
        plt.savefig("Idiff.png")


        I1, I2  = getIs(head1, comps1, comps2, R, theta)


        plt.clf()
        plt.plot(R, I1)
        plt.plot(R, I2)
        plt.xscale("log")
        plt.xlabel("Rad")
        plt.ylabel("I")
        plt.grid(True)
        plt.minorticks_on()
        plt.savefig("BulgeRad.png")




    #computing bulge radius
    try:
        rbulge = newton(getDiffx, 0, args=(head1, comps1, comps2, theta)) 
    except:
        print("solution not found for initial parameter")
        rbulge=0

    return rbulge, N1, N2, theta 



def getDiff(head1, comps1, comps2, R, theta):


    Idiff = np.array([])

    for r in R:

        Ir1 = GetIr().Ir(head1, comps1, r, theta)
        Ir2 = GetIr().Ir(head1, comps2, r, theta)

        Ird = Ir1 - Ir2
        Idiff = np.append(Idiff, Ird)



    return Idiff  



def getIs(head1, comps1, comps2, R, theta):


    I1 = np.array([])
    I2 = np.array([])

    for r in R:

        Ir1 = GetIr().Ir(head1, comps1, r, theta)
        Ir2 = GetIr().Ir(head1, comps2, r, theta)

        I1 = np.append(I1, Ir1)
        I2 = np.append(I2, Ir2)



    return I1, I2




def getDiffx(r, head1, comps1, comps2, theta):


    Ir1 = GetIr().Ir(head1, comps1, r, theta)
    Ir2 = GetIr().Ir(head1, comps2, r, theta)

    Irdx = Ir1 - Ir2


    return Irdx





class GetIr:


    def Ir(self, head, comps, R, theta):


        #comps.Rad = comps.Rad*head.scale
        comps.Flux = 10**((head.mgzpt - comps.Mag)/2.5)

        k = gammaincinv(2*comps.Exp, 0.5)

        denom1 = (2*np.pi*comps.Rad**2)*(np.exp(k))
        denom2 = (comps.Exp)*(k**(-2*comps.Exp))
        #denom3 = (gamma(2*comps.Exp))*(comps.AxRat) 
        denom3 = (gamma(2*comps.Exp))

        denom = denom1*denom2*denom3 
        
        comps.Ie = comps.Flux/denom

        maskgal = (comps.Active == True) 


        Itotr = self.Itotser(R, comps.Ie[maskgal], comps.Rad[maskgal], comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], theta) 

        #me = -2.5*np.log10(Itotr)

        return Itotr 

     
    def Itotser(self, R: float, Ie: list, rad: list, n: list, q: list, pa: list, theta: float) -> float:

        ItotR = self.Iser(R, Ie, rad, n, q, pa, theta) 

        return ItotR.sum()



    def Iser(self, R: float, Ie: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''sersic flux to a determined R'''


        k = gammaincinv(2*n, 0.5)

        Rcor = GetRadAng(R, q, pa, theta) 

        Ir = Ie*np.exp(-k*((Rcor/Re)**(1/n) - 1))

        
        return Ir














#!/usr/bin/env python3


import numpy as np
from astropy.io import fits
import sys
import os

def imarith(ImageFile: str, output: str, image2: str, add: float, mul: float, div: float, sub: float)-> None: 

    if not os.path.exists(ImageFile):
        print ('{} image filename does not exist!'.format(ImageFile))
        sys.exit()

    if image2:
        if not os.path.exists(image2):
            print ('{} image filename does not exist!'.format(image2))
            sys.exit()

        hdu2 = fits.open(image2)
        dataImage2 = hdu2[0].data




    # original file
    hdu = fits.open(ImageFile)
    dataImage = hdu[0].data


    if add:
        if image2:
            newdata = dataImage + dataImage2
        else:
            newdata = dataImage + add

    if mul:
        if image2:
            newdata = dataImage*dataImage2
        else:
            newdata = dataImage*mul

    if div:
        if image2:
            newdata = dataImage/dataImage2
        else:
            newdata = dataImage/div

    if sub:
        if image2:
            newdata = dataImage - dataImage2
        else:
            newdata = dataImage - sub



    hdu[0].data = newdata 
    hdu.writeto(output, overwrite=True)


    hdu.close()

    if image2:
        hdu2.close()








from astropy.io import fits
import numpy as np

import os 

from scipy.optimize import bisect

import copy
from scipy.special import gamma, gammainc, gammaincinv

class GalHead():
    '''store the header of galfit file'''

    inputimage = "none.fits"
    outimage = "none-out.fits"
    sigimage = "none"
    psfimage = "none"
    psfsamp = 1
    maskimage = "none"
    constraints = "none"
    xmin = 0
    xmax = 1
    ymin = 0
    ymax = 1
    convx = 1
    convy = 1
    mgzpt = 25
    scale = 1
    scaley = 1
    display = "regular"
    P = 0

    #extra info  
    imgidx = "sci"
    flagidx = False
    num = 1
    flagnum = False
    exptime = 1
    tempmask = "tempmask.fits"

### class for Galfit components
class GalComps:
    '''stores the components of galfit file'''

    N = np.array([])
    NameComp = np.array([])        #0)
    PosX = np.array([])            #1)   
    PosY = np.array([])            #2)   
    Mag = np.array([])             #3)
    Rad = np.array([])             #4)
    Exp = np.array([])             #5)
    Exp2 = np.array([])            #6)  for moffat
    Exp3 = np.array([])            #7)  for moffat
                                   #8)  There is No 8 in any galfit model
    AxRat = np.array([])           #9)  AxisRatio
    PosAng = np.array([])          #10) position angle
    skip = np.array([])            #z)  skip model

    Active = np.array([])            #activate component  for galaxy

    # store the flags related to parameters
    PosXFree = np.array([])            #1)   
    PosYFree = np.array([])            #2)   
    MagFree = np.array([])             #3)
    RadFree = np.array([])             #4)
    ExpFree = np.array([])             #5)
    Exp2Free = np.array([])            #6)  for moffat
    Exp3Free = np.array([])            #7)  for moffat
                                   #8)  There is No 8 in any galfit model
    AxRatFree = np.array([])           #9)  AxisRatio
    PosAngFree = np.array([])          #10) position angle

    # computed parameters:
    Rad50 = np.array([])
    SerInd = np.array([])
    Rad50kpc = np.array([])
    Rad50sec = np.array([])
    Rad90 = np.array([])
    AbsMag = np.array([])
    Lum = np.array([])
    Flux = np.array([])
    PerLight = np.array([])
    me = np.array([])
    mme = np.array([])
    kser = np.array([])


    KronRad = np.array([])
    PetRad = np.array([])





class GalSky:
    '''stores the value of the GALFIT file'''

    sky = 0 
    dskyx = 0
    dskyy = 0
    skip = 0 

    skyfree = 1 
    dskyxfree = 0 
    dskyyfree = 0 
 

class DataImg:

    img = np.array([[1,1],[1,1]])
    model = np.array([[1,1],[1,1]])
    imres = np.array([[1,1],[1,1]])

    mask = np.array([[1,1],[1,1]])
    sigma = np.array([[1,1],[1,1]])
    imsnr = np.array([[1,1],[1,1]])
    imchi = np.array([[1,1],[1,1]])
    impsf = np.array([[1,1],[1,1]])


def readDataImg(ellconf, galhead):

    dataimg = DataImg()

    #reading galaxy and model images from file
    errmsg="file {} does not exist".format(galhead.outimage)

    assert os.path.isfile(galhead.outimage), errmsg

    if ellconf.flagmodel == False:
        # hdu 1 => image   hdu 2 => model
        hdu = fits.open(galhead.outimage)
        dataimg.img = (hdu[1].data.copy()).astype(float)
        dataimg.model = (hdu[2].data.copy()).astype(float)
        dataimg.imres = (hdu[3].data.copy()).astype(float)
        hdu.close()

    else:
        hdu = fits.open(galhead.inputimage)

        i = 0 #index of data 
        if galhead.flagidx:
            if galhead.flagnum:
                dataimg.img = (hdu[galhead.imgidx, galhead.num].data).astype(float)
            else:    
                dataimg.img = (hdu[galhead.imgidx].data).astype(float)
        else:
            dataimg.img = (hdu[i].data).astype(float)

        hdu.close()

        hdu = fits.open(ellconf.inputmodel)
        dataimg.model = (hdu[i].data).astype(float)
        hdu.close()

        dataimg.imres = galpar.img - galpar.model


    ### reading mask image from file

    if galhead.tempmask is not None:

        errmsg="file {} does not exist".format(galhead.tempmask)
        assert os.path.isfile(galhead.tempmask), errmsg

        i = 0 #index of data 
        if ellconf.flagmodel == False:
            hdu = fits.open(galhead.tempmask)
            mask = hdu[i].data
            dataimg.mask=np.array(mask,dtype=bool)
            hdu.close()
        else:
            hdu = fits.open(galhead.maskimage)
            mask = hdu[i].data
            dataimg.mask=np.array(mask,dtype=bool)
            hdu.close()

    else:
        dataimg.mask = None
 

    return dataimg



class Galfit():

    def __init__(self, File: str):
        self.File = File


    def ReadHead(self) -> GalHead:
        '''reads header of galfit file'''
        inputf = self.File 

        galhead = GalHead() # class for header

        maskimage = ""
        #    skylevel=0

        GalfitFile = open(inputf,"r")

        # All lines including the blank ones
        lines = (line.rstrip() for line in GalfitFile)
        lines = (line.split('#', 1)[0] for line in lines)  # remove comments
        # remove lines containing only comments
        lines = (line.rstrip() for line in lines)
        lines = (line for line in lines if line)  # Non-blank lines

        lines = list(lines)
        index = 0

        while index < len(lines):

            #================================================================================
            # IMAGE and GALFIT CONTROL PARAMETERS
            #A) tempfits/A2399-3-2.fits      # Input data image (FITS file)
            #B) A2399-215615.96-m073822.7-337-out.fits      # Output data image block
            #C) tempfits/none-3-2.fits      # Sigma image name (made from data if blank or "none")
            #D) psfs/PSF-1309-721.fits          # Input PSF image and (optional) diffusion kernel
            #E) 1                   # PSF fine sampling factor relative to data
            #F) mask-337            # Bad pixel mask (FITS image or ASCII coord list)
            #G) constraints         # File with parameter constraints (ASCII file)
            #H) 129  809  265  809  # Image region to fit (xmin xmax ymin ymax)
            #I) 60     60           # Size of the convolution box (x y)
            #J) 21.672              # Magnitude photometric zeropoint
            #K) 0.680  0.680        # Plate scale (dx dy)   [arcsec per pixel]
            #O) regular             # Display type (regular, curses, both)
            #P) 0                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps

            line = lines[index]
            (tmp) = line.split()

            if len(tmp) > 1: #avoids empty options

                if tmp[0] == "A)":     # input image
                    galhead.inputimage=tmp[1]

                if tmp[0] == "B)":     # out image
                    galhead.outimage = tmp[1]

                if tmp[0] == "C)":   # sigma 
                    galhead.sigimage = tmp[1]

                if tmp[0] == "D)":  # psf file 
                    galhead.psfimage = tmp[1]

                if tmp[0] == "E)":  #psf sampling 
                    galhead.psfsamp = int(tmp[1])

                if tmp[0] == "F)":     # mask image
                    try:
                        galhead.maskimage = tmp[1]
                    except IndexError:
                        galhead.maskimage = "None"

                if tmp[0] == "G)":  # psf file 
                    galhead.constraints = tmp[1]


                if tmp[0] == "H)":     # region fit box
                    galhead.xmin = int(tmp[1])
                    galhead.xmax = int(tmp[2])
                    galhead.ymin = int(tmp[3])
                    galhead.ymax = int(tmp[4])

                if tmp[0] == "I)":     # convolution size 
                    galhead.convx = int(tmp[1])
                    try:
                        galhead.convy = int(tmp[2])
                    except:
                        galhead.convy = galhead.convx 

                if tmp[0] == "J)":     # mgzpt
                    galhead.mgzpt = float(tmp[1])

                if tmp[0] == "K)":     # plate scale
                    galhead.scale = float(tmp[1])
                    try:
                        galhead.scaley = float(tmp[2])
                    except:
                        galhead.scaley = galhead.scale 

                if tmp[0] == "O)":     # display 
                    galhead.display = tmp[1]

                if tmp[0] == "P)":     # optional output 
                    galhead.P = int(tmp[1])


            index += 1


        ###### check for extension in name #####
        chars = set('[]') 
        numbers=set('1234567890')
        if any((c in chars) for c in galhead.inputimage): 
            print("Ext Found") 
            galhead.flagidx=True
            (filename,imgidxc) = galhead.inputimage.split("[")
            (imgidx,trash)=imgidxc.split("]")

            if any((n in numbers) for n in imgidx):
                galhead.flagnum=True
                (imgidx,num)=imgidx.split(",")
                num=int(num)

            galhead.inputimage=filename
            galhead.imgidx=imgidx
            galhead.num=num
      
        ####################

        return galhead


    def ReadComps(self) -> GalComps:
        '''reads all the components in the galfit file'''

        File = self.File

        galcomps = GalComps()
         
        GalfitFile = open(File,"r")

        # All lines including the blank ones
        lines = (line.rstrip() for line in GalfitFile)
        lines = (line.split('#', 1)[0] for line in lines)  # remove comments
        # remove lines containing only comments
        lines = (line.rstrip() for line in lines)
        lines = (line for line in lines if line)  # Non-blank lines

        lines = list(lines)
        index = 0

        N = 0


        while index < len(lines):

            line = lines[index]
            (tmp) = line.split()

            #init values
            NameComp = "none"
            PosX = 0
            PosY = 0
            Mag = 99
            Rad = 0
            Exp = 0
            Exp2 = 0
            Exp3 = 0
            AxRat = 1
            PosAng = 0
            skip = 0

            flagcomp = False 

            Xfree = 1
            Yfree = 1
            Magfree = 1
            Radfree = 1
            Expfree = 1
            Exp2free = 1
            Exp3free = 1
            AxRatfree = 1
            PosAngfree = 1



            if (tmp[0] == "0)") and (tmp[1] != 'sky'):

                namec = tmp[1] 
                N = N + 1 
                NameComp = namec
                while (tmp[0] != "Z)"):

                    index += 1
                    line = lines[index]
                    (tmp) = line.split()

                    if tmp[0] == "1)":   # center
                        PosX = float(tmp[1])
                        PosY = float(tmp[2])
                        Xfree = int(tmp[3])
                        Yfree = int(tmp[4])
                    if tmp[0] == "3)" :    # mag 
                        Mag = float(tmp[1])
                        Magfree = int(tmp[2])
                    if tmp[0] == "4)" :
                        Rad = float(tmp[1])
                        Radfree = int(tmp[2])
                    if tmp[0] == "5)" : 
                        Exp = float(tmp[1])
                        Expfree = int(tmp[2])
                    if tmp[0] == "6)": 
                        Exp2 = float(tmp[1])
                        try:
                            Exp2free = int(tmp[2])
                        except:
                            Exp2free = 0 
                    if tmp[0] == "7)":  
                        Exp3 = float(tmp[1])
                        try:
                            Exp3free = int(tmp[2])
                        except:
                            Exp3free = 0 
                    if tmp[0] == "9)":
                        AxRat = float(tmp[1])
                        try:
                            AxRatfree = int(tmp[2])
                        except:
                            AxRatfree = 0 
                    if tmp[0] == "10)":
                        PosAng = float(tmp[1])
                        PosAngfree = int(tmp[2])
                    if tmp[0] == "Z)":
                        skip=int(tmp[1])

                galcomps.PosX = np.append(galcomps.PosX, PosX)
                galcomps.PosY = np.append(galcomps.PosY, PosY)
                galcomps.NameComp = np.append(galcomps.NameComp, NameComp)
                galcomps.N = np.append(galcomps.N, N)
                    
                galcomps.Mag = np.append(galcomps.Mag, Mag)
                galcomps.Rad = np.append(galcomps.Rad, Rad)
                galcomps.Exp = np.append(galcomps.Exp, Exp)
                galcomps.Exp2 = np.append(galcomps.Exp2, Exp2)
                galcomps.Exp3 = np.append(galcomps.Exp3, Exp3)
                galcomps.AxRat = np.append(galcomps.AxRat, AxRat)
                galcomps.PosAng = np.append(galcomps.PosAng, PosAng)
                galcomps.skip = np.append(galcomps.skip, skip)
                galcomps.Active = np.append(galcomps.Active, flagcomp)


                galcomps.PosXFree = np.append(galcomps.PosXFree, Xfree)
                galcomps.PosYFree = np.append(galcomps.PosYFree, Yfree)
                galcomps.MagFree = np.append(galcomps.MagFree, Magfree)
                galcomps.RadFree = np.append(galcomps.RadFree, Radfree)
                galcomps.ExpFree = np.append(galcomps.ExpFree, Expfree)
                galcomps.Exp2Free = np.append(galcomps.Exp2Free, Exp2free)
                galcomps.Exp3Free = np.append(galcomps.Exp3Free, Exp3free)
                galcomps.AxRatFree = np.append(galcomps.AxRatFree, AxRatfree)
                galcomps.PosAngFree = np.append(galcomps.PosAngFree, PosAngfree)



            index += 1

        GalfitFile.close()

        #filling the rest of arrays with zeros: 
        arsiz = len(galcomps.N)

        galcomps.Rad50 = np.zeros(arsiz)
        galcomps.SerInd = np.zeros(arsiz)
        galcomps.Rad50kpc = np.zeros(arsiz)
        galcomps.Rad50sec = np.zeros(arsiz)
        galcomps.Rad90 = np.zeros(arsiz)
        galcomps.AbsMag = np.zeros(arsiz)
        galcomps.Lum = np.zeros(arsiz)
        galcomps.Flux = np.zeros(arsiz)
        galcomps.PerLight = np.zeros(arsiz)
        galcomps.me = np.zeros(arsiz)
        galcomps.mme = np.zeros(arsiz)
        galcomps.kser = np.zeros(arsiz)
        galcomps.KronRad = np.zeros(arsiz)
        galcomps.PetRad = np.zeros(arsiz)



        return galcomps 


    def ReadSky(self) -> GalSky:
        '''reads the sky value of the galfit file'''

        File = self.File    

        galsky = GalSky()

        GalfitFile = open(File,"r")

        # All lines including the blank ones
        lines = (line.rstrip() for line in GalfitFile)
        lines = (line.split('#', 1)[0] for line in lines)  # remove comments
        # remove lines containing only comments
        lines = (line.rstrip() for line in lines)
        lines = (line for line in lines if line)  # Non-blank lines

        lines = list(lines)
        index = 0

        N = 0


        while index < len(lines):

            line = lines[index]
            (tmp) = line.split()

            #init values
            NameComp = "sky"
            sky = 0
            dskyx = 0
            dskyy = 0
            skip = 0

            skyfree = 1
            dskyxfree = 0
            dskyyfree = 0


            if (tmp[0] == "0)") and (tmp[1] == NameComp):

                namec = tmp[1] 

                while (tmp[0] != "Z)"):

                    index += 1
                    line = lines[index]
                    (tmp) = line.split()

                    if tmp[0] == "1)":   # center
                        sky = float(tmp[1])
                        skyfree = float(tmp[2])
                    if tmp[0] == "2)" :    # mag 
                        dskyx = float(tmp[1])
                        dskyxfree = int(tmp[2])
                    if tmp[0] == "3)" :
                        dskyy = float(tmp[1])
                        dskyyfree = int(tmp[2])
                    if tmp[0] == "Z)":
                        skip=int(tmp[1])

                #note if this work in this way or using append
                galsky.NameComp =  NameComp
                    
                galsky.sky = sky
                galsky.dskyx =  dskyx
                galsky.dskyy =  dskyy
                galsky.skip =  skip

                galsky.skyfree =  skyfree
                galsky.dskyxfree = dskyxfree
                galsky.dskyyfree =  dskyyfree
     

            index += 1

        GalfitFile.close()

     
        return galsky


def numParFree(galcomps: GalComps) -> int:
    '''obtains the number of free parameters. This function does NOT 
    count the sky as a free param'''



    p1 = 0
    p2 = 0
    p3 = 0
    p4 = 0
    p5 = 0
    p6 = 0
    p7 = 0
    p8 = 0
    p9 = 0


    parmask1 = (galcomps.Active == True) & (galcomps.PosXFree == 1) 
    parmask2 = (galcomps.Active == True) & (galcomps.PosYFree == 1) 
    parmask3 = (galcomps.Active == True) & (galcomps.MagFree == 1) 
    parmask4 = (galcomps.Active == True) & (galcomps.RadFree == 1) 
    parmask5 = (galcomps.Active == True) & (galcomps.ExpFree == 1) 
    parmask6 = (galcomps.Active == True) & (galcomps.Exp2Free == 1) 
    parmask7 = (galcomps.Active == True) & (galcomps.Exp3Free == 1) 
    parmask8 = (galcomps.Active == True) & (galcomps.AxRatFree == 1) 
    parmask9 = (galcomps.Active == True) & (galcomps.PosAngFree == 1) 

    if parmask1.any():
        p1 = np.sum(galcomps.PosXFree[parmask1])
    if parmask2.any():
        p2 = np.sum(galcomps.PosYFree[parmask2])
    if parmask3.any():
        p3 = np.sum(galcomps.MagFree[parmask3])
    if parmask4.any():
        p4 = np.sum(galcomps.RadFree[parmask4])
    if parmask5.any():
        p5 = np.sum(galcomps.ExpFree[parmask5])
    if parmask6.any():
        p6 = np.sum(galcomps.Exp2Free[parmask6])
    if parmask7.any():
        p7 = np.sum(galcomps.Exp3Free[parmask7])
    if parmask8.any():
        p8 = np.sum(galcomps.AxRatFree[parmask8])
    if parmask9.any():
        p9 = np.sum(galcomps.PosAngFree[parmask9])


    pt = p1 + p2 + p3 + p4 + p5 + p6 + p7 + p8 + p9


    return int(pt)


def numParSkyFree(galsky: GalSky) -> int:
    '''obtains the number of free parameters for sky function'''


    p1 = 0
    p2 = 0
    p3 = 0

    parmask1 =  galsky.skyfree == 1 
    parmask2 =  galsky.dskyxfree == 1 
    parmask3 =  galsky.dskyyfree == 1 

    if parmask1:
        p1 = 1 
    if parmask2:
        p2 = 1 
    if parmask3:
        p3 = 1 

    pt = p1 + p2 + p3 


    return int(pt)





def numComps(galcomps: GalComps, name: str) -> int:
    '''obtains the number of components'''


    if name == 'all':
        nummask = (galcomps.Active == True) & ( (galcomps.NameComp == 'sersic') 
                    | (galcomps.NameComp == 'expdisk') | (galcomps.NameComp == 'gaussian')
                    | (galcomps.NameComp == 'devauc'))

    else:
        nummask = (galcomps.Active == True) & (galcomps.NameComp == name)




    N = galcomps.Active[nummask].size

    return N





def SelectGal(galcomps: GalComps, distmax: float, n_comp: int) -> GalComps:
    '''changes Flag to true for those components who belongs
        to the same galaxy of n_comp'''

    galcomps.Active.fill(False)

    idx = np.where(galcomps.N ==  n_comp)

    assert idx[0].size !=0, 'component not found'

    n_idx = idx[0].item(0)

    if distmax > 10:
        print("Warning: maximum distance among components is greater than 10")
        print("Equations' solutions only apply for components that share the same center")


    galcomps.Active[n_idx] = True #main component

    posx = galcomps.PosX[n_idx] 
    posy = galcomps.PosY[n_idx]      

    dx = galcomps.PosX - posx
    dy = galcomps.PosY - posy

    dist = np.sqrt((dx)**2 + (dy)**2)

    maskdist = (dist <= distmax ) 

    galcomps.Active[maskdist] = True

    return galcomps



### Sersic components 
class GetReff:
    '''class to obtain the effective radius for the whole galaxy'''

    def GetReSer(self, galhead: GalHead, comps: GalComps, eff: float, theta: float) -> float:


        maskgal = (comps.Active == True) 

        comps.Flux = 10**((galhead.mgzpt - comps.Mag)/2.5)

        totFlux = comps.Flux[maskgal].sum()

        totmag = -2.5*np.log10(totFlux) + galhead.mgzpt

        a = 0.1
        b = comps.Rad[maskgal][-1] * 1000  # hope it doesn't crash

        Reff = self.solveSerRe(a, b, comps.Flux[maskgal], comps.Rad[maskgal], 
                comps.Exp[maskgal], comps.AxRat[maskgal], comps.PosAng[maskgal], totFlux, eff, theta)


        return Reff, totmag



    def solveSerRe(self, a: float, b: float, flux: list, rad: list, n: list, q: list, pa: list, totFlux: float, eff: float, theta: float) -> float:
        "return the Re of a set of Sersic functions. It uses Bisection"

        try:

            Re = bisect(self.funReSer, a, b, args=(flux, rad, n, q, pa, totFlux, eff, theta))
        except:
            print("solution not found for the given range")
            Re = 0

        return Re


    def funReSer(self, R: float, flux: list, rad: list, n: list, q: list, pa: list, totFlux: float, eff: float, theta: float) -> float:
        

        fun = self.Ftotser(R, flux, rad, n, q, pa, theta) - totFlux*eff

        return fun
     
    def Ftotser(self, R: float, flux: list, rad: list, n: list, q: list, pa: list, theta: float) -> float:

        ftotR = self.Fser(R, flux, rad, n, q, pa, theta) 

        return ftotR.sum()



    def Fser(self, R: float, Flux: list, Re: list, n: list, q: list, pa: list, theta: float) -> float:
        '''sersic flux to a determined R'''
        
        k = gammaincinv(2*n, 0.5)

        Rcor = GetRadAng(R, q, pa, theta) 

        X = k*(Rcor/Re)**(1/n) 

        Fr = Flux*gammainc(2*n, X) ##esta funcion esta mal 
        
        return Fr



def conver2Sersic(galcomps: GalComps) -> GalComps:
    ''' function to convert exponential, gaussian params to Sersic params'''

    comps =  copy.deepcopy(galcomps)

    maskdev = comps.NameComp == "devauc"
    maskexp = comps.NameComp == "expdisk"
    maskgas = comps.NameComp == "gaussian"


    K_GAUSS = 0.6931471805599455 #constant k for gaussian
    K_EXP = 1.6783469900166612 # constant k for expdisk
    SQ2 = np.sqrt(2) 

    #for gaussian functions
    if maskgas.any():
        comps.Exp[maskgas] = 0.5 
        comps.Rad[maskgas] = comps.Rad[maskgas]/2.354 #converting to sigma 
        comps.Rad[maskgas] = SQ2*(K_GAUSS**0.5)*comps.Rad[maskgas] #converting to Re 


    #for de vaucouleurs
    if maskdev.any():
        comps.Exp[maskdev] = 4

    #for exponential disks
    if maskexp.any():
        comps.Exp[maskexp] = 1
        comps.Rad[maskexp] = K_EXP*comps.Rad[maskexp] #converting to Re


    return comps



def GetRadAng(R: float, q: list, pa: list, theta: float) -> float:
    '''Given an ellipse and an angle it returns the radius in angle direction. 
    Theta are the values for the galaxy and the others for every component'''



    #changing measured angle from y-axis to x-axis
    # and changing to rads:
    newpa = (pa + 90)*np.pi/180 #angle of every component
    theta = (theta + 90)*np.pi/180 #angle of direction of R 

    #bim = q * R

    ecc = np.sqrt(1 - q**2)

    alpha = theta - newpa #this is the direction 


    bell =  R*np.sqrt(1 - (ecc*np.cos(alpha))**2)


    aell = bell/q  #rad to evalue for every component



    return aell 




def galfitLastFit(directory: str) -> str:


    # Directory containing the files
    #directory = "." #actual directory

    # List all files in the directory
    files = os.listdir(directory)

    # Filter files that match the prefix 'galfit.'
    galfit_files = [f for f in files if f.startswith('galfit.') and f[7:].isdigit()]

    # Extract the numerical suffix and find the maximum
    max_file = max(galfit_files, key=lambda x: int(x[7:]))


    return max_file

#print galfit files
def galPrintHeader(hdl, galhead: GalHead)-> bool:
    "print GALFIT header in a file"


    A = galhead.inputimage
    B = galhead.outimage
    C = galhead.sigimage 
    D = galhead.psfimage 
    E = galhead.psfsamp
    F = galhead.maskimage
    G = galhead.constraints
    xlo = galhead.xmin
    xhi = galhead.xmax
    ylo = galhead.ymin
    yhi = galhead.ymax
    convx = galhead.convx
    convy = galhead.convy
    J = galhead.mgzpt
    platedx = galhead.scale
    platedy = galhead.scaley 
    O = galhead.display
    P = galhead.P
    #S = 0 


    # k Check
    # print to filehandle
    # the header for GALFIT

    lineZ = "==================================================================================================\n"
    lineX = "# IMAGE PARAMETERS \n"
    lineA = "A) {}    # Input Data image (FITS file)                            \n".format(A)
    lineB = "B) {}    # Output data image block                                 \n".format(B)
    lineC = "C) {}    # Sigma image name (made from data if blank or \"none\")  \n".format(C)
    lineD = "D) {}    # Input PSF image and (optional) diffusion kernel         \n".format(D)
    lineE = "E) {}    # PSF fine sampling factor relative to data               \n".format(E)
    lineF = "F) {}    # Bad pixel mask (FITS image or ASCII coord list)         \n".format(F)
    lineG = "G) {}    # File with parameter constraints (ASCII file)            \n".format(G)
    lineH = "H) {} {} {} {}   # Image region to fit (xmin xmax ymin ymax)               \n".format(xlo, xhi, ylo, yhi)
    lineI = "I) {} {}  # Size of the convolution box (x y)                       \n".format(convx, convy)
    lineJ = "J) {}     # Magnitude photometric zeropoint                         \n".format(J)
    lineK = "K) {} {}  # Plate scale (dx dy). \[arcsec per pixel\]               \n".format(platedx, platedy)
    lineO = "O) {}     # Display type (regular, curses, both)                    \n".format(O)
    lineP = "P) {}     # Choose 0=optimize, 1=model, 2=imgblock, 3=subcomps      \n".format(P)
    #lineS = "S) {}     # Modify/create objects interactively?                    \n".format(S)
    lineY = " \n"

    line0 = "# INITIAL FITTING PARAMETERS                                                     \n"
    line1 = "# \n"
    line2 = "#   For object type, allowed functions are:                                      \n"
    line3 = "#       nuker, sersic, expdisk, devauc, king, psf, gaussian, moffat,             \n"
    line4 = "#       ferrer, powsersic, sky, and isophote.                                    \n"
    line5 = "# \n"
    line6 = "#  Hidden parameters will only appear when they're specified:                    \n"
    line7 = "#      C0 (diskyness/boxyness),                                                  \n"
    line8 = "#      Fn (n=integer, Azimuthal Fourier Modes),                                  \n"
    line9 = "#      R0-R10 (PA rotation, for creating spiral structures).                     \n"
    line10 = "# \n"

    line11 = "# column 1:  Parameter number                                                               \n"
    line12 = "# column 2:                                                                                 \n"
    line13 = "#          -- Parameter 0:    the allowed functions are: sersic, nuker, expdisk             \n"
    line14 = "#                             edgedisk, devauc, king, moffat, gaussian, ferrer, psf, sky    \n"
    line15 = "#          -- Parameter 1-10: value of the initial parameters                               \n"
    line16 = "#          -- Parameter C0:   For diskiness/boxiness                                        \n"
    line17 = "#                             <0 = disky                                                    \n"
    line18 = "#                             >0 = boxy                                                     \n"
    line19 = "#          -- Parameter Z:    Outputting image options, the options are:                    \n"
    line20 = "#                             0 = normal, i.e. subtract final model from the data to create \n"
    line21 = "#                             the residual image                                            \n"
    line22 = "#                             1 = Leave in the model -- do not subtract from the data       \n"
    line23 = "#                                                                                           \n"
    line24 = "# column 3: allow parameter to vary (yes = 1, no = 0)                                       \n"
    line25 = "# column 4: comment                                                                         \n"
    line26 = " \n"

    line27 = "==================================================================================================\n"

    hdl.write(lineZ)
    hdl.write(lineX)
    hdl.write(lineA)
    hdl.write(lineB)
    hdl.write(lineC)
    hdl.write(lineD)
    hdl.write(lineE)
    hdl.write(lineF)
    hdl.write(lineG)
    hdl.write(lineH)
    hdl.write(lineI)
    hdl.write(lineJ)
    hdl.write(lineK)
    hdl.write(lineO)
    hdl.write(lineP)
    #hdl.write(lineS)
    hdl.write(lineY)

    hdl.write(line0)
    hdl.write(line1)
    hdl.write(line2)
    hdl.write(line3)
    hdl.write(line4)
    hdl.write(line5)
    hdl.write(line6)
    hdl.write(line7)
    hdl.write(line8)
    hdl.write(line9)
    hdl.write(line10)

    hdl.write(line11)
    hdl.write(line12)
    hdl.write(line13)
    hdl.write(line14)
    hdl.write(line15)
    hdl.write(line16)
    hdl.write(line17)
    hdl.write(line18)
    hdl.write(line19)
    hdl.write(line20)
    hdl.write(line21)
    hdl.write(line22)
    hdl.write(line23)
    hdl.write(line24)
    hdl.write(line25)
    hdl.write(line26)
    hdl.write(line27)

    return True


def galPrintComp(hdl: str, ncomp: int, idx: int, galcomps: GalComps)-> bool:
    "print GALFIT Sersic function to filehandle"
    # k Check

    # print to filehandle

    line00 = "# Object number: {}   \n".format(ncomp)
    line01 = " 0)   {}   #  Object type      \n".format(galcomps.NameComp[idx])
    line02 = " 1) {:.2f}  {:.2f}  {}  {}  #  position x, y  [pixel] \n".format(
        galcomps.PosX[idx], galcomps.PosY[idx], galcomps.PosXFree[idx], galcomps.PosYFree[idx])
    line03 = " 3) {:.2f}  {}    #  total magnitude  \n".format(
        galcomps.Mag[idx], galcomps.MagFree[idx])
    line04 = " 4) {:.2f}  {}    #  R_e     [Pixels] \n".format(
            galcomps.Rad[idx], galcomps.RadFree[idx])
    line05 = " 5) {}     {}   #  Sersic exponent (deVauc=4, expdisk=1) \n".format(
        galcomps.Exp[idx], galcomps.ExpFree[idx])
    line06 = " 6)  {}  {}      #  ---------------- \n".format(galcomps.Exp2[idx],
                galcomps.Exp2Free[idx])
    line07 = " 7)  {}  {}      #  ---------------- \n".format(galcomps.Exp3[idx], 
            galcomps.Exp3Free[idx])
    line08 = " 8)  0.0000       0   #  ----------------                                \n"
    line09 = " 9) {:.2f}   {}   #  axis ratio (b/a)  \n".format(
        galcomps.AxRat[idx], galcomps.AxRatFree[idx])
    line10 = "10) {:.2f}    {}  #  position angle (PA)  [Degrees: Up=0, Left=90]   \n".format(
        galcomps.PosAng[idx], galcomps.PosAngFree[idx])
    lineZ = " Z) {}         #  Skip this model in output image?  (yes=1, no=0) \n".format(
        galcomps.skip[idx])
    line11 = "\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)
    hdl.write(line08)
    hdl.write(line09)
    hdl.write(line10)
    hdl.write(lineZ)
    hdl.write(line11)

    return True




def galPrintSky(hdl: str , ncomp: int, galsky: GalSky) -> bool:
    "Print GALFIT sky function to filehandle using  GalSky class"


    line00 = "# Object number: {}                 \n".format(ncomp)
    line01 = " 0)      sky            #    Object type                                        \n"
    line02 = " 1) {:.2f}   {}   # sky background        [ADU counts]  \n".format(galsky.sky, galsky.skyfree)
    line03 = " 2) {:.2f}  {}    # dsky/dx (sky gradient in x) \n".format(galsky.dskyx,galsky.dskyxfree)
    line04 = " 3) {:.2f}  {}    # dsky/dy (sky gradient in y) \n".format(galsky.dskyy,galsky.dskyyfree)
    line05 = " Z) {}           # Skip this model in output image?  (yes=1, no=0) \n".format(galsky.skip)
    line06 = "\n"
    line07 = "================================================================================\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)

    return True





#! /usr/bin/env python3

import numpy as np
import sys
import os
import subprocess as sp
from astropy.io import fits
import os.path  
import scipy
import scipy.special
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx


from scipy import stats
from matplotlib import gridspec

from astropy.io.votable import parse
from mgefit.sectors_photometry import sectors_photometry
from mgefit.find_galaxy import find_galaxy

from matplotlib.ticker import (MultipleLocator, FormatStrFormatter,NullFormatter,
                               AutoMinorLocator,LogLocator,LinearLocator,AutoLocator)


from galfitools.galin.MaskDs9 import GetAxis 

from galfitools.mge.mge2galfit import GetInfoEllip, Ds9ell2Kronellv2, GetSize, GetPmax 

import argparse



def sbProf(args):
    '''creates a surface brightness profile'''

    image = args.Image
    ds9reg = args.Ds9Region
    mgzpt = args.mgzpt
    mask =  args.mask
    axrat = args.axrat
    angle = args.angle
    sky = args.sky
    plate = args.plate
    output = args.output
    center = args.center
    ranx = args.ranx
    rany = args.rany
    logx = args.logx
    pix = args.pix
    grid = args.grid

    rad = args.rad
    rad2 = args.rad2

    #configuration

    conf = Config()

    conf.image = image  
    conf.ds9reg = ds9reg
    conf.mgzpt = mgzpt
    conf.mask = mask
    conf.skylevel = sky
    conf.scale = plate
    conf.center = center
    conf.output = output
    conf.ranx = ranx
    conf.rany = rany
    conf.pix = pix
    conf.grid = grid 
    conf.logx = logx 
    conf.rad = rad
    conf.rad2 = rad2



    dataimg = readDataImg(conf)

    #hdu = fits.open(image)
    #img = hdu[0].data
    #img=img.astype(float)


    conf.exptime = GetExpTime(image)

    (ncol, nrow) = GetAxis(image)
 

    obj, xpos, ypos, rx, ry, angle2 = GetInfoEllip(conf.ds9reg)
    xx, yy, Rkron, theta, eps = Ds9ell2Kronellv2(xpos,ypos,rx,ry,angle2)

    if center:
        print('center of ds9 ellipse region will be used')
        xpeak, ypeak = xpos, ypos
    else:        
        (xmin, xmax, ymin, ymax) = GetSize(xx, yy, Rkron, theta, eps, ncol, nrow)
        xpeak, ypeak = GetPmax(dataimg.img, dataimg.mask, xmin, xmax, ymin, ymax)




    # I have to switch x and y coordinates, don't ask me why

    #xpeak, ypeak = ypeak, xpeak




    conf.xc = xpeak
    conf.yc = ypeak 

    if angle:
        conf.parg = angle 
    else:
        conf.parg =  theta

    if axrat:
        conf.qarg = axrat
        conf.eps = 1 - axrat
    else:
        conf.eps = eps
        conf.qarg = 1 - eps



    print("galaxy found at ", xpeak + 1, ypeak + 1)
    print("Axis ratio {:.2f}, Angle = {:.2f}".format(conf.qarg, conf.parg))

    print("Sky = ", sky)



    # removing background from galaxy and model images 
    dataimg.img = dataimg.img - conf.skylevel
    #dataimg.model = dataimg.model - conf.skylevel



    #   numsectors=19 default
    numsectors = conf.sectors

    minlevel = conf.minlevel  # minimun value for sky

    #call to sectors_photometry for galaxy and model

    sectgalax = SectPhot(conf, dataimg, n_sectors = numsectors, minlevel = minlevel)
           


    print("creating plots..")

    limx, limy = EllipSectors(conf, sectgalax, n_sectors = numsectors)
                                

    print("plot file: ", conf.output)
 


    ##############################################
    ##############################################
    ##############################################

    if conf.dplot:
        plt.pause(1.5)
    plt.savefig(conf.output, dpi = conf.dpival)


class DataImg:

    img = np.empty([2,2])

    mask = np.empty([2,2])


def readDataImg(conf):

    dataimg = DataImg()

    #reading galaxy and model images from file
    errmsg="file {} does not exist".format(conf.image)

    assert os.path.isfile(conf.image), errmsg

        # hdu 1 => image   hdu 2 => model
    hdu = fits.open(conf.image)
    dataimg.img = (hdu[0].data.copy()).astype(float)
    hdu.close()


    ### reading mask image from file

    if conf.mask:

        errmsg="file {} does not exist".format(conf.mask)
        assert os.path.isfile(conf.mask), errmsg

        hdu = fits.open(conf.mask)
        mask = hdu[0].data
        dataimg.mask=np.array(mask,dtype=bool)
        hdu.close()

    else:
        dataimg.mask = np.array([])

 


    return dataimg


def SectPhot(conf, dataimg, n_sectors = 19, minlevel = 0):
    """ calls to function sectors_photometry for galaxy and model """


    maskb = dataimg.mask

    eps = 1 - conf.qarg

    #if ellconf.dplot:
    plt.clf()
    print("")

    ############
    # I have to switch x and y values because they are different axes for
    # numpy:
    #if ellconf.flagmodel == False:
    yctemp = conf.xc
    xctemp = conf.yc
    #else:
    #    yctemp = ellconf.inxc
    #    xctemp = ellconf.inyc


    # and angle is different as well:
    #angsec = 90 - conf.parg
    angsec=conf.parg

    ###################################################
    #if fit == 'gal':
    #  galaxy:

    if conf.mask:
        sectimg = sectors_photometry(dataimg.img, eps, angsec, xctemp, yctemp, minlevel = minlevel,
                        plot = conf.dplot, badpixels = maskb, n_sectors = n_sectors)
    else:
        sectimg = sectors_photometry(dataimg.img, eps, angsec, xctemp, yctemp, minlevel = minlevel,
                        plot = conf.dplot, n_sectors = n_sectors)
 


    #if ellconf.dplot:
    plt.pause(1)  # Allow plot to appear on the screen
    plt.savefig(conf.namesec)


    #if fit == 'mod':
    #  model: 
    #    sectimg = sectors_photometry(dataimg.model, eps, angsec, xctemp, yctemp, minlevel=0,
    #            plot = ellconf.dplot, badpixels = maskb, n_sectors = n_sectors)


    #    if ellconf.dplot:
    #        plt.pause(1)  # Allow plot to appear on the screen
    #        plt.savefig(ellconf.namemod)



    return sectimg



def EllipSectors(conf, sectgalax, n_sectors = 19, minlevel = 0):


   
    #galaxy
    xradq, ysbq, ysberrq = sect2xy(sectgalax, conf,  n_sectors)


    # Plotting
    limx,limy, axsec = PlotSB(xradq, ysbq, ysberrq, conf, conf.scale)



    axsec.legend(loc=1)

    return limx,limy


def sect2xy(sect, conf, n_sectors):

    #######################################
    #######################################
    # model
    stidx = np.argsort(sect.radius)

    mgerad = sect.radius[stidx]
    mgecount = sect.counts[stidx]
    mgeangle = sect.angle[stidx]
    mgeanrad = np.deg2rad(mgeangle)

    ab = conf.qarg

    aellab = mgerad * np.sqrt((np.sin(mgeanrad)**2)/ab**2 + np.cos(mgeanrad)**2)


    aellarc = aellab*conf.scale

    # formula according to cappellary mge manual
    mgesb = conf.mgzpt - 2.5*np.log10(mgecount/conf.exptime) \
            + 2.5*np.log10(conf.scale**2) + 0.1 #- ellconf.Aext

    stidxq = np.argsort(aellarc)

    xarc = aellarc[stidxq]
    ymge = mgesb[stidxq]


    ######  Function to order SB along X-axis for model

    xrad, ysb, ysberr = FindSB(xarc, ymge, n_sectors)


    return xrad, ysb, ysberr


def FindSB(xarcq, ymgeq, numsectors):
    # the xarcq array must be ordered
    # use mag instead of counts

    xradq=[]
    ysbq=[]
    ysberrq=[]
    xradq=np.array(xradq)
    ysbq=np.array(ysbq)
    ysberrq=np.array(ysberrq)

    numsave=0
    tot=xarcq.size
    count=0
    for i in range(tot,0,-1):

        lima=i-numsectors
        limb=i

        if xarcq[lima:limb].size == 0:
            break
        else:
            valstd=np.std(xarcq[lima:limb])
            if valstd < 0.1:
                numsave=count
                break
            count=count+1

    init=numsave%numsectors
    n=init

    num=np.int32((xarcq.size-init)/numsectors)
    n=xarcq.size-init
    for i in range(num,0,-1):

        lima=n-numsectors
        limb=n

        xradq=np.append(xradq,np.mean(xarcq[lima:limb]))
        ysbq=np.append(ysbq,np.mean(ymgeq[lima:limb]))
        #ysberrq=np.append(ysberrq,np.std(ymgeq[lima:limb]))
        ysberrq=np.append(ysberrq,stats.sem(ymgeq[lima:limb])) #standard error is more appropiate

        n=n-numsectors

    return xradq, ysbq, ysberrq


def PlotSB(xradq,ysbq,ysberrq, conf, scale):
    """  Produces final best-fitting plot  """

    # subplot for arc sec axis
    plt.close('all')


    #ULISES begin
    #fig, (axsec,axred) = plt.subplots(2, sharex=True, sharey=False)
    #fig, axsec = plt.subplots(1)
    #gs = gridspec.GridSpec(2, 1,height_ratios=[3,1])
    #gs.update(hspace=0.07)
    #ULISES end 


    fig, axsec = plt.subplots()

    if conf.ranx:
        (xmin, xmax) = conf.ranx[0], conf.ranx[1]

    if conf.rany:
        (ymin, ymax) = conf.rany[0], conf.rany[1]

    minrad = np.min(xradq)
    maxrad = np.max(xradq)

    mincnt = np.min(ysbq)
    maxcnt = np.max(ysbq)
    xran = minrad * (maxrad/minrad)**np.array([-0.02, +1.02])
    yran = mincnt * (maxcnt/mincnt)**np.array([+1.05, -0.05]) #inverted axis

   
    # ULISES begin
    #axsec = plt.subplot(gs[0])

    #axsec.set_ylabel(r"Surface Brightness $(mag\; arcsec^{-2})$")
    # ULISES end

    axsec.set_ylabel(r"Surface Brightness $(mag\; arcsec^{-2})$")

    axsec.errorbar(xradq, ysbq,yerr=ysberrq,fmt='o-',capsize=2,color='blue',markersize=0.7,label="galaxy",linewidth=2)


    #if ellconf.flagalax == False:
    #    axsec.errorbar(xradm, ysbm,yerr=ysberrm,fmt='o-',capsize=2,color='blue',markersize=0.7,label="Model",linewidth=2)


    if conf.rany == True:
        axsec.set_ylim(ymax,ymin) #inverted
    else:
        axsec.set_ylim(yran)

    if conf.logx == True:

        axsec.set_xscale("log")

        locmaj = LogLocator(base=10,numticks=12)
        axsec.xaxis.set_major_locator(locmaj)

        locmin = LogLocator(base=10.0,subs=(0.2,0.4,0.6,0.8),numticks=12)
        axsec.xaxis.set_minor_locator(locmin)
        axsec.xaxis.set_minor_formatter(NullFormatter())

    else:
        axsec.xaxis.set_minor_locator(AutoMinorLocator())
        axsec.xaxis.set_major_locator(AutoLocator())


    axsec.tick_params(which='both', width=2)
    axsec.tick_params(which='major', length=7)
    axsec.tick_params(which='minor', length=4, color='r')


    #begin psf fwhm 
    #if ellconf.flagfwhm: 
    #    xpos = ellconf.fwhm*scale
    #    axsec.axvline(x=xpos,  linestyle='--', color='k', linewidth=2)
    # end 


    # ULISES begin
    #axsec.axes.xaxis.set_ticklabels([])
    # ULISES end 


    #axsec.axes.xaxis.set_ticklabels([])

    axsec.yaxis.set_minor_locator(AutoMinorLocator())
    axsec.yaxis.set_major_locator(AutoLocator())


    # ULISES begin
    # Residual plot
    #if len(ysbq) < len(ysbm):
    #    ysbm = ysbm[len(ysbm)-len(ysbq):]
    
    #elif len(ysbq) > len(ysbm):
    #    ysbq = ysbq[len(ysbq)-len(ysbm):]
    #    ysberrq = ysberrq[len(ysberrq)-len(ysbq):]

    #if len(ysbq) < len(ysberrm):
    #    ysberrm = ysberrm[len(ysberrm)-len(ysbq):]
    #elif len(ysbq) > len(ysberrm):
    #    ysbq = ysbq[len(ysbq)-len(ysberrm):]        

    #residual = ((ysbq-ysbm)/ysbq)*100 # (data-model)/data in percentage

    #err = ((ysbm/ysbq**2)**2) * ysberrq**2 + ((1/ysbq)**2) * ysberrm**2 
    #err = np.sqrt(err)*100
    #axred = plt.subplot(gs[1])

    #if len(xradq) != len(residual):
    #    axred.errorbar(xradm, residual, yerr=err,fmt='.',capsize=2,color='k')
    #else:
    #    axred.errorbar(xradq, residual, yerr=err,fmt='.',capsize=2,color='k')

    #axred.axhline(y=0,ls='dashed', color='k')
    axsec.set_xlabel('Radius $(arcsec)$')
    #axred.set_ylabel('Residual (%)')
    #axred.set_ylim(-2,2)
    # ULISES end


    #if conf.flaglogx == True:

    #    axred.set_xscale("log")

    #    locmaj = LogLocator(base=10,numticks=12)
    #    axred.xaxis.set_major_locator(locmaj)

    #    locmin = LogLocator(base=10.0,subs=(0.2,0.4,0.6,0.8),numticks=12)
    #    axred.xaxis.set_minor_locator(locmin)
    #    axred.xaxis.set_minor_formatter(NullFormatter())
    #else:
    #    axred.xaxis.set_minor_locator(AutoMinorLocator())
    #    axred.xaxis.set_major_locator(AutoLocator())



    #axred.tick_params(which='both', width=2)
    #axred.tick_params(which='major', length=7)
    #axred.tick_params(which='minor', length=4, color='r')



    if conf.ranx:
        axsec.set_xlim(xmin,xmax)
    #  axred.set_xlim(xmin,xmax) #ulises plot
    else:
        axsec.set_xlim(xran)
      # axred.set_xlim(xran) #ulises plot
     



    if conf.pix == True:

        axpix = axsec.twiny()

        axpix.set_xlabel("Pixels")
        x1, x2 = axsec.get_xlim()

        axpix.set_xlim(x1/scale, x2/scale)

        axpix.figure.canvas.draw()


        if conf.logx == True:
            axpix.set_xscale("log")
            locmaj = LogLocator(base=10,numticks=12)
            axpix.xaxis.set_major_locator(locmaj)

            locmin = LogLocator(base=10.0,subs=(0.2,0.4,0.6,0.8),numticks=12)
            axpix.xaxis.set_minor_locator(locmin)
            axpix.xaxis.set_minor_formatter(NullFormatter())
        else:
            axpix.xaxis.set_minor_locator(AutoMinorLocator())
            axpix.xaxis.set_major_locator(AutoLocator())

        axpix.tick_params(which='both', width=2)
        axpix.tick_params(which='major', length=7)
        axpix.tick_params(which='minor', length=4, color='r')

        axret=axsec
    else:
        axret=axsec

    if conf.grid == True:
        # Customize the major grid
        axsec.grid(which='major', linestyle='-', linewidth='0.7', color='black')
        # Customize the minor grid
        axsec.grid(which='minor', linestyle=':', linewidth='0.5', color='black')


    #change the linewidth of the axis
    for axis in ['top','bottom','left','right']:
        axsec.spines[axis].set_linewidth(1.5)
        #axred.spines[axis].set_linewidth(1.5)

    if conf.rad:

        axsec.axvline(x=conf.rad,  linestyle='--', color='m', linewidth=2)

    if conf.rad2:

        axsec.axvline(x=conf.rad2,  linestyle='--', color='c', linewidth=2)




    return xran,yran,axret



## class for parameters
class Config:


    logx = False
    image = "none.fits" 

    mgzpt = 25
    ds9reg = "none.reg"
    mask = 'mask.fits'
    scale = 1 
    center = False
    exptime = 1

    namesec = "gal.png"

    #init
    xc = 1 
    yc = 1 

    ranx = [0,0]
    rany = [0,0]
 
    qarg=1
    parg=0

    dpival=150

    minlevel=0
    sectors=19


    image = "none.fits"


    #output file
    output = "out.png"

    pix = False
    grid = False


    # sky parameters:
    skylevel = 0


    dplot=True

    Aext = 0  # surface brightness correction for plots

    rad = 0
    rad2 = 0



def GetExpTime(Image):
    # k Check
    "Get exposition time from the image"

    try:
        hdu = fits.open(Image)
        exptime = hdu[0].header["EXPTIME"]
        hdu.close()
    except: 
        exptime = 1
    return float(exptime)


if __name__ == '__main__':

    mainSbProf()



#! /usr/bin/env python3


import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from astropy.io import fits
from scipy import stats
from matplotlib.patches import Ellipse

from astropy import units as U

import argparse




def displayCube(cubeimage: str, namecube: str, dpival: int, brightness: float, contrast:float, 
                    cmap: str, scale: float, noplot: bool) -> None:


    ######################
    #shows the image cube#
    ######################{


    ell=[]


    ShowCube(cubeimage, namepng = namecube, dpival 
             = dpival, bri = brightness, con = contrast, 
              cmap = cmap, ellipse = ell, plate = scale)


    if noplot is False:
        plt.pause(1.5)
 
    plt.close()

    #}
    #####################


def Comp2Ellip(galhead, galcomps, N, lw=1):
    ''' converts galfit component parameter into an Ellipse object''' 


    ellipses = [] 

    #color value
    values = range(N)
    jet = cm = plt.get_cmap('jet') 
    cNorm  = colors.Normalize(vmin=0, vmax=values[-1])
    scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=jet)

    
    for idx, item in enumerate(galcomps.N):

        if galcomps.Active[idx] == True:
            # correcting coordinates
            xc = galcomps.PosX[idx] - galhead.xmin + 1
            yc = galcomps.PosY[idx] - galhead.ymin + 1


            pa = galcomps.PosAng[idx] + 90

            w = galcomps.Rad[idx]
            h = galcomps.Rad[idx]*galcomps.AxRat[idx]


            colorVal = scalarMap.to_rgba(values[idx])


            ell=Ellipse((xc, yc), width = w, height = h, angle = pa,
                         edgecolor = colorVal,
                         facecolor = 'none',
                         linewidth = lw)

            ellipses.append(ell)


    return ellipses






class ShowCube:

    def __init__(self, cubeimg: str, namepng="cubeout.png", dpival=100, 
                bri = 0, con = 1, cmap='viridis', ellipse=[], plate = 1):
        """
        This routine shows the GALFIT output cube image: galaxy, model and residual    
        """

        
        hdu = fits.open(cubeimg)
        data = (hdu[1].data.copy()).astype(float)
        model = (hdu[2].data.copy()).astype(float)
        residual = (hdu[3].data.copy()).astype(float)
        hdu.close()

        flatmodimg = model.flatten()  
        flatresimg = residual.flatten()  

        flatmodimg.sort()
        flatresimg.sort()

        restot = len(flatresimg)

        restop = round(.9*restot)
        resbot = round(.1*restot)

        modtot = len(flatmodimg)

        modtop = round(.9*modtot)
        modbot = round(.1*modtot)



        modimgpatch = flatmodimg#[modbot:modtop]
        resimgpatch = flatresimg[resbot:restop]

        resmin = np.min(resimgpatch)
        resmax = np.max(resimgpatch)


        modmin = np.min(modimgpatch)
        modmax = np.max(modimgpatch)


        data = data.clip(modmax/1e4,modmax)
        model = model.clip(modmax/1e4)

        modmin = modmax/1e4


        middle = (modmax - modmin)/2


        #brightness auto-adjust according to the contrast value 

        Autobri = middle*(con -1) + modmin*(1-con) 


        #user can re-adjust brightness in addition to Autobri
        newdata = con*(data - middle) + middle + Autobri + bri*(modmax-middle)
        newmodel = con*(model - middle) + middle + Autobri + bri*(modmax-middle)



        mask=data < 0 
        data[mask] = 1 # avoids problems in log
     
        fig, (ax1, ax2, ax3) = plt.subplots(figsize=(14, 5), nrows = 1, ncols = 3)

        im = ax1.imshow(newdata, origin ='lower', interpolation='nearest', norm 
                    = colors.LogNorm(vmin=modmin, vmax=modmax), cmap = cmap)


        ax1.set_title('Data')

        y,x = data.shape

        xt = .02*x
        yt = .02*y

        lxline = round(.1*x)
        lyline = round(.1*y)
        x1 = [xt, xt+lxline]
        y1 = [lyline, lyline]

        arcsec = lxline*plate*U.arcsec 


        if arcsec.value >= 60:
            lxlinearc = arcsec.to("arcmin").value
            s = "{}\'".format(round(lxlinearc))
        else:
            lxlinearc = arcsec.value
            s = "{}\'\'".format(round(lxlinearc))
 
        ax1.plot(x1, y1, color="white", linewidth=3)

        ax1.text(xt+round(lxline/5),lyline+yt,s,color='white',fontsize=14)
    
        #ax1.set_xlabel(r'$\circ$')
        #ax2.set_ylabel('23\"')

        for ell in ellipse:
            ax1.add_patch(ell)


        ax2.imshow(newmodel, origin='lower', interpolation='nearest', norm 
                    = colors.LogNorm(vmin = modmin, vmax = modmax), cmap = cmap)


        ax2.set_title('GALFIT Model')

        ax3.imshow(residual, origin='lower', vmin = resmin, vmax = resmax, cmap = cmap)
        ax3.set_title('Residual')

        fig.subplots_adjust(left=0.04, right=0.98, bottom=0.02, top=0.98)
        plt.savefig(namepng, dpi = dpival)
    


#end of program
if __name__ == '__main__':
    mainShowCube()



#! /usr/bin/env python

import numpy as np
import sys
import os
import subprocess as sp
from astropy.io import fits
import os.path
import scipy
import argparse

from galfitools.galin.MaskDs9 import GetAxis


def makeMask(sexfile: str, image: str, maskfile: str, scale: float, satfileout: str) -> None:



    sexarsort   = "sexsort.cat"

    satscale = 1
    satoffset = 0


    print ("Creating masks....\n")

    (NCol, NRow) = GetAxis(image)

    Total = CatArSort(sexfile,scale,sexarsort,NCol,NRow)

    print ("Creating sat region files....\n")

    if not os.path.exists(satfileout):
        print("Saturation file not found. Creating one")
        satfileout  = "ds9sat.reg"
        ds9satbox(satfileout,sexfile,satscale,satoffset) # crea archivo  Saturacion reg


    ##### segmentation mask

    MakeImage(maskfile, NCol, NRow)

    MakeMask(maskfile, sexarsort, scale, 0, satfileout)  # offset set to 0
    MakeSatBox(maskfile, satfileout, Total + 1, NCol, NRow) #make sat region




def ds9satbox (satfileout,output,satscale,satoffset):
    "Creates a file for ds9 which selects bad saturated regions"

    scaleflag=1
    offsetflag=1
    regfileflag=1
    magflag=1
    clasflag=1

    flagsat=4      ## flag value when object is saturated (or close to)
    maxflag=128    ## max value for flag
    check=0
    regflag = 0    ## flag for saturaded regions




    f_out = open(satfileout, "w")

    N,Alpha,Delta,X,Y,Mg,Kr,Fluxr,Isoa,Ai,E,Theta,Bkgd,Idx,Flg=np.genfromtxt(output,delimiter="",unpack=True)


    line="image \n"
    f_out.write(line)



    for idx, item in enumerate(N):


        bi=Ai[idx]*(1-E[idx])

        Theta[idx] = Theta[idx] * np.pi /180  #rads!!!


        Rkronx = satscale * 2 * Ai[idx] * Kr[idx]  + satoffset
        Rkrony = satscale * 2 * bi * Kr[idx]  + satoffset



        if Rkronx == 0:
            Rkronx = 1

        if Rkrony == 0:
            Rkrony = 1

        check=CheckFlag(Flg[idx],flagsat,maxflag)  ## check if object has saturated regions

        if (check):

            line="box({0},{1},{2},{3},0) # color=red move=0 \n".format(X[idx],Y[idx],Rkronx,Rkrony)
            f_out.write(line)

            line2="point({0},{1}) # point=boxcircle font=\"times 10 bold\" text={{ {2} }} \n".format(X[idx],Y[idx],N[idx])
            f_out.write(line2)


    f_out.close()





def MakeMask(maskimage, catfile, scale, offset, regfile):
    "Create a mask image using ellipses for every Object of catfile. Now includes offset"
    # k Check

    checkflag = 0
    flagsat = 4  # flag value when object is saturated (or close to)
    maxflag = 128  # max value for flag

    regflag = 0  # flag for saturaded regions

    n, alpha, delta, xx, yy, mg, kr, fluxrad, ia, ai, e, theta, bkgd, idx, flg, sxmin, sxmax, symin, symax, sxsmin, sxsmax, sysmin, sysmax = np.genfromtxt(
        catfile, delimiter="", unpack=True)

    n = n.astype(int)
    flg = flg.astype(int)

    #print("Creating Masks for sky \n")

    Rkron = scale * ai * kr + offset

    mask = Rkron < 1
    if mask.any():
        Rkron[mask] = 1

    
    i = 0 #index indicated where the data is located

    hdu = fits.open(maskimage)
    img = hdu[i].data



    print ("Creating ellipse masks for every object \n")

    for idx, val in enumerate(n):

        # check if object doesn't has saturaded regions
        checkflag = CheckFlag(flg[idx], flagsat, maxflag)
        # check if object is inside of a saturaded box region indicated by
        # user in ds9
        regflag = CheckSatReg(xx[idx], yy[idx], regfile, Rkron[idx], theta[idx], e[idx])

        if (checkflag == False) and (regflag == False):

            #print ("Creating ellipse mask for object {}  \n".format(n[idx]))
            img = MakeKron(img, n[idx], xx[idx], yy[idx], Rkron[idx], theta[idx], e[
                idx], sxsmin[idx], sxsmax[idx], sysmin[idx], sysmax[idx])

        #elif(checkflag == True or regflag == True):

            

    print ("ignoring objects where one or more pixels are saturated \n")

    hdu[i].data = img
    hdu.writeto(maskimage, overwrite=True)
    hdu.close()

    return True


def MakeKron(imagemat, idn, x, y, R, theta, ell, xmin, xmax, ymin, ymax):
    "This subroutine create a Kron ellipse within a box defined by: xmin, xmax, ymin, ymax"

    # Check

    xmin = int(xmin)
    xmax = int(xmax)
    ymin = int(ymin)
    ymax = int(ymax)

    q = (1 - ell)
    bim = q * R

    theta = theta * np.pi / 180  # Rads!!!

    ypos, xpos = np.mgrid[ymin - 1:ymax, xmin - 1:xmax]

    dx = xpos - x
    dy = ypos - y

    landa = np.arctan2(dy, dx)

    mask = landa < 0
    if mask.any():
        landa[mask] = landa[mask] + 2 * np.pi

    landa = landa - theta

    angle = np.arctan2(np.sin(landa) / bim, np.cos(landa) / R)

    xell = x + R * np.cos(angle) * np.cos(theta) - bim * \
        np.sin(angle) * np.sin(theta)
    yell = y + R * np.cos(angle) * np.sin(theta) + bim * \
        np.sin(angle) * np.cos(theta)

    dell = np.sqrt((xell - x)**2 + (yell - y)**2)
    dist = np.sqrt(dx**2 + dy**2)

    mask = dist < dell
    imagemat[ypos[mask], xpos[mask]] = idn

    return imagemat


def MakeSatBox(maskimage, region, val, ncol, nrow):
    "Create a mask for saturated regions"
    "Regions must be in DS9 box regions format"

    # k Check

    #	fileflag=1
    i = 0 # index where data is located 
    hdu = fits.open(maskimage)
    img = hdu[i].data

    with open(region) as f_in:

        next(f_in)

        # All lines including the blank ones
        lines = (line.rstrip() for line in f_in)
        lines = (line.split('#', 1)[0] for line in lines)  # remove comments
        # remove lines containing only comments
        lines = (line.rstrip() for line in lines)
        lines = (line for line in lines if line)  # Non-blank lines

        for line in lines:

            words = line.split(" ")
            if (words[0] != "image" and words[0] != "physical" and words[0] != "global"):

                (box, info) = line.split('(')

                if(box == "box"):

                    (xpos, ypos, xlong, ylong, trash) = info.split(',')

                    xpos = float(xpos)
                    ypos = float(ypos)
                    xlong = float(xlong)
                    ylong = float(ylong)

                    xlo = (xpos - xlong / 2)
                    xhi = (xpos + xlong / 2)

                    ylo = (ypos - ylong / 2)
                    yhi = (ypos + ylong / 2)

                    xlo = int(xlo)
                    xhi = int(xhi)

                    ylo = int(ylo)
                    yhi = int(yhi)

                    if (xlo < 1):

                        xlo = 1

                    if (xhi > ncol):

                        xhi = ncol

                    if (ylo < 1):

                        ylo = 1

                    if (yhi > nrow):

                        yhi = nrow

                    img[ylo - 1:yhi, xlo - 1:xhi] = val

    hdu[i].data = img
    hdu.writeto(maskimage, overwrite=True)
    hdu.close()

    return True




def MakeImage(newfits, sizex, sizey):
    "create a new blank Image"
    # k Check
    if os.path.isfile(newfits):
        print("{} deleted; a new one is created \n".format(newfits))
    hdu = fits.PrimaryHDU()
    hdu.data = np.zeros((sizey, sizex))
    hdu.writeto(newfits, overwrite=True)

    return True



def CatArSort(SexCat,scale,SexArSort,NCol,NRow):
    # k Check

    # sort the sextractor
    # catalog by magnitude,
    # get sizes for objects
    # and write it in a new file

    # The sextractor catalog must contain the following parameters:
    #   1 NUMBER                 Running object number
    #   2 ALPHA_J2000            Right ascension of barycenter (J2000)                      [deg]
    #   3 DELTA_J2000            Declination of barycenter (J2000)                          [deg]
    #   4 X_IMAGE                Object position along x                                    [pixel]
    #   5 Y_IMAGE                Object position along y                                    [pixel]
    #   6 MAG_APER               Fixed aperture magnitude vector                            [mag]
    #   7 KRON_RADIUS            Kron apertures in units of A or B
    #   8 FLUX_RADIUS            Fraction-of-light radii                                    [pixel]
    #   9 ISOAREA_IMAGE          Isophotal area above Analysis threshold                    [pixel**2]
    #  10 A_IMAGE                Profile RMS along major axis                               [pixel]
    #  11 ELLIPTICITY            1 - B_IMAGE/A_IMAGE
    #  12 THETA_IMAGE            Position angle (CCW/x)                                     [deg]
    #  13 BACKGROUND             Background at centroid position                            [count]
    #  14 CLASS_STAR             S/G classifier output
    #  15 FLAGS                  Extraction flags


    print("Sorting and getting sizes for objects \n")

    n, alpha, delta, xx, yy, mg, kr, fluxrad, ia, ai, e, theta, bkgd, idx, flg = np.genfromtxt(
        SexCat, delimiter="", unpack=True)

    n = n.astype(int)
    flg = flg.astype(int)

#    ai = ai.astype(float)
#    kr = kr.astype(float)

#    scale = scale.astype(float)


    Rkron = scale * ai * kr

    Rwsky = scale * ai * kr + 10  + 20


#   considering to use only  KronScale instead of SkyScale
#    Rwsky = parvar.KronScale * ai * kr + parvar.Offset + parvar.SkyWidth

    Bim = (1 - e) * Rkron

    Area = np.pi * Rkron * Bim *(-1)

    (sxmin, sxmax, symin, symax) = GetSize(xx, yy, Rkron, theta, e, NCol, NRow)

    (sxsmin, sxsmax, sysmin, sysmax) = GetSize(xx, yy, Rwsky, theta, e, NCol,NRow)

    f_out = open(SexArSort, "w")

    index = Area.argsort()
    for i in index:

        line = "{} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {}\n".format(n[i], alpha[i], delta[i], xx[i], yy[i], mg[i], kr[i], fluxrad[i], ia[i], ai[i], e[i], theta[i], bkgd[i], idx[i], flg[i], int(
            np.round(sxmin[i])), int(np.round(sxmax[i])), int(np.round(symin[i])), int(np.round(symax[i])), int(np.round(sxsmin[i])), int(np.round(sxsmax[i])), int(np.round(sysmin[i])), int(np.round(sysmax[i])))

        f_out.write(line)

    f_out.close()

    return len(n)


def GetSize(x, y, R, theta, ell, ncol, nrow):
    "this subroutine get the maximun"
    "and minimim pixels for Kron and sky ellipse"
    # k Check
    q = (1 - ell)
    bim = q * R

    theta = theta * (np.pi / 180)  # rads!!

    # getting size

    xmin = x - np.sqrt((R**2) * (np.cos(theta))**2 +
                       (bim**2) * (np.sin(theta))**2)

    xmax = x + np.sqrt((R**2) * (np.cos(theta))**2 +
                       (bim**2) * (np.sin(theta))**2)

    ymin = y - np.sqrt((R**2) * (np.sin(theta))**2 +
                       (bim**2) * (np.cos(theta))**2)

    ymax = y + np.sqrt((R**2) * (np.sin(theta))**2 +
                       (bim**2) * (np.cos(theta))**2)


    mask = xmin < 1
    if mask.any():
        if isinstance(xmin,np.ndarray):
            xmin[mask] = 1
        else:
            xmin = 1

    mask = xmax > ncol

    if mask.any():
        if isinstance(xmax,np.ndarray):
            xmax[mask] = ncol
        else:
            xmax = ncol

    mask = ymin < 1
    if mask.any():
        if isinstance(ymin,np.ndarray):
            ymin[mask] = 1
        else:
            ymin = 1

    mask = ymax > nrow
    if mask.any():
        if isinstance(ymax,np.ndarray):
            ymax[mask] = nrow
        else:
            ymax = nrow


    return (xmin, xmax, ymin, ymax)

def CheckFlag(val,check,max):
   "Check for flag contained in $val, returns 1 if found "

   flag = False
   mod = 1

   while (mod != 0):


       res = int(val/max)

       if (max == check and res == 1 ):

           flag=True


       mod = val % max

       val = mod
       max = max/2



   return flag


def CheckSatReg(x,y,filein,R,theta,ell):
   "Check if object is inside of saturated region. returns True if at least one pixel is inside"
    ## saturaded region as indicated by ds9 box region
    ## returns 1 if object center is in saturaded region


   q = (1 - ell)

   bim = q * R

   theta = theta * np.pi /180  ## Rads!!!

   flag = False
   fileflag =1



   with open(filein) as f_in:

       lines = (line.rstrip() for line in f_in) # All lines including the blank ones
       lines = (line.split('#', 1)[0] for line in lines) # remove comments
       lines = (line.rstrip() for line in lines)   # remove lines containing only comments
       lines = (line for line in lines if line) # Non-blank lines

       for line in lines:

           words = line.split(" ")
           if (words[0] != "image" and words[0] != "physical" and words[0] != "global"):

               (box,info)=line.split('(')

               if(box == "box"):

                   (xpos,ypos,xlong,ylong,trash)=info.split(',')

                   xpos=float(xpos)
                   ypos=float(ypos)
                   xlong=float(xlong)
                   ylong=float(ylong)


                   xlo = xpos - xlong/2
                   xhi = xpos + xlong/2

                   ylo = ypos - ylong/2
                   yhi = ypos + ylong/2

                   dx = xpos - x
                   dy = ypos - y

                   landa=np.arctan2( dy,dx )

                   if landa < 0:
                       landa=landa + 2 * np.pi


                   landa = landa - theta

                   angle = np.arctan2(np.sin(landa)/bim, np.cos(landa)/R)

                   xell =  x + R * np.cos(angle)* np.cos(theta)  - bim * np.sin(angle) * np.sin(theta)
                   yell =  y + R * np.cos(angle)* np.sin(theta)  + bim * np.sin(angle) * np.cos(theta)


                   if ( (xell > xlo and xell < xhi) and (yell > ylo and yell < yhi)  ):

                       flag=True
                       break


   return flag













#end of program
if __name__ == '__main__':
    mainMakeMask()



#! /usr/bin/env python

import numpy as np
import argparse
import os
from astropy.io import fits
import subprocess as sp
import scipy
import sys
import copy

from scipy.special import gamma, gammainc, gammaincinv


from galfitools.galin.galfit import Galfit, conver2Sersic, SelectGal, numComps, GetRadAng


from galfitools.galin.MaskDs9 import GetAxis

from galfitools.galin.galfit import GalComps, GalHead
from galfitools.galin.galfit import numParFree, numParSkyFree



class HeadInfo():
    '''store the flags of the header of the galfit file'''

    inputimageflag =  False   
    outimageflag = False 
    sigimageflag = False 
    psfimageflag = False 
    maskimageflag = False 
    constraintsflag = False 
    xsizeflag = False 
    ysizeflag = False
    convxflag = False 
    convyflag = False 


    inputimage = "none.fits"
    outimage = "none-out.fits"
    sigimage = "none"
    psfimage = "none"
    psfsamp = 1
    maskimage = "none"
    constraints = "none"
    xmin = 0
    xmax = 1
    ymin = 0
    ymax = 1
    convx = 1
    convy = 1
    mgzpt = 25
    scale = 1
    scaley = 1
    display = "regular"
    P = 0







def checkFile(galfitFile: str, dis: int) -> float:
    '''prints information of the galfit File '''




    galfit = Galfit(galfitFile)

    head = galfit.ReadHead()
    galcomps = galfit.ReadComps()

    galsky = galfit.ReadSky()

    headinfo  = HeadInfo()

    #copying info from header

    headinfo = copy2info(head, headinfo)


    galcomps = SelectGal(galcomps, dis, 1) #modify functions such that 0 will clean galcomps


    #================================================================================
    # IMAGE and GALFIT CONTROL PARAMETERS
    #A) A407.fits           # Input data image (FITS file)
    #B) A407-G1-mge.fits      # Output data image block
    #C) None                # Sigma image name (made from data if blank or "none") 
    #D) psf.fits            # Input PSF image and (optional) diffusion kernel
    #E) 1                   # PSF fine sampling factor relative to data 
    #F) mask.fits           # Bad pixel mask (FITS image or ASCII coord list)
    #G) cons.G1G2           # File with parameter constraints (ASCII file) 
    #H) 613  1290 882  1275 # Image region to fit (xmin xmax ymin ymax)
    #I) 71     71           # Size of the convolution box (x y)
    #J) 21.4710             # Magnitude photometric zeropoint 
    #K) 0.6800  0.6800      # Plate scale (dx dy)   [arcsec per pixel]
    #O) regular             # Display type (regular, curses, both)
    #P) 0                   # Choose: 0=optimize, 1=model, 2=imgblock, 3=subcomps


    if os.path.exists(headinfo.inputimage):
        headinfo.inputimageflag = True

    if os.path.exists(headinfo.outimage):
        headinfo.outimageflag = True

    if (os.path.exists(headinfo.sigimage)) or (headinfo.sigimage == 'none') :
            headinfo.sigimageflag = True

    if (os.path.exists(headinfo.psfimage)) or (headinfo.psfimage == 'none') :
            headinfo.psfimageflag = True
    
    if (os.path.exists(headinfo.maskimage)) or (headinfo.maskimage == 'none'):
            headinfo.maskimageflag = True

    if (os.path.exists(headinfo.constraints)) or (headinfo.constraints == 'none'):
            headinfo.constraintsflag = True

    if headinfo.xmax > headinfo.xmin:
        headinfo.xsizeflag = True

    if headinfo.ymax > headinfo.ymin:
        headinfo.ysizeflag = True


    if headinfo.psfimageflag:

        (nx, ny) = GetAxis(headinfo.psfimage)
        

        if headinfo.convx > nx:
            headinfo.convxflag = True

        if headinfo.convy > ny:
            headinfo.convyflag = True

    

    galax = np.zeros(len(galcomps.N))
    mag = np.zeros(len(galcomps.N))

    mag.fill(99)


    mag =np.copy(galcomps.Mag) 
    cont=1

    dmax = dis

    #counting components per galaxy
    for idx, item in enumerate(galcomps.N):
         dx = (galcomps.PosX[idx] - galcomps.PosX)
         dy = (galcomps.PosY[idx] - galcomps.PosY)
         d = np.sqrt(dx**2 + dy**2)
         maskdis = (d <= dmax) & (galax == 0)
         if maskdis.any():
             galax[maskdis] = cont
             cont=cont+1
     
 

    galcomps.Active = True
    freepar = numParFree(galcomps) #computing the number of free parameters
    freeparsky = numParSkyFree(galsky) #computing the number of free parameters

    freepar = freepar + freeparsky
   

    return headinfo, galax, mag, freepar


def copy2info(head, head2):

    head2.inputimage = head.inputimage
    head2.outimage = head.outimage 
    head2.sigimage = head.sigimage 
    head2.psfimage = head.psfimage 
    head2.psfsamp = head.psfsamp
    head2.maskimage =  head.maskimage  
    head2.constraints = head.constraints 
    head2.xmin = head.xmin
    head2.xmax =  head.xmax
    head2.ymin =  head.ymin
    head2.ymax =  head.ymax
    head2.convx =  head.convx
    head2.convy = head.convy
    head2.mgzpt = head.mgzpt
    head2.scale = head.scale 
    head2.scaley = head.scaley
    head2.display = head.display
    head2.P =  head.P





    return head2


#############################################################################
######################### End of program  ###################################
#     ______________________________________________________________________
#    /___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/_/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/
##############################################################################
if __name__ == '__main__':
  main()






#!/usr/bin/env python3

import numpy as np
import sys
import os
import stat
import subprocess as sp
import os.path
from astropy.io import fits
import scipy
import scipy.special
import matplotlib.pyplot as plt


import argparse

from galfitools.galin.MaskDs9 import GetAxis 
# computes sky using GALFIT
#change it and use skyRem instead

def galfitSky(imgname, maskfile, mgzpt, scale, X, Y, sky)-> None:
    

    flagpos=False




    (NCol, NRow) = GetAxis(imgname)



    convbox=100

    #    sky=0.55
    fit=1
    Z=0

    xpos = X
    ypos = Y
    anglegass=35.8


    #    scale = 0.0455  # arcsec/pixel

    ###################################################
    #  Create GALFIT input file header to compute sky #
    ###################################################

    #    imgname = "ngc4342.fits"
    root_ext = os.path.splitext(imgname)
    TNAM = root_ext[0]
 

    exptime = GetExpTime(imgname)

    ##################
    #################

    # Here we use an accurate four gaussians MGE PSF for
    # the HST/WFPC2/F814W filter, taken from Table 3 of
    # Cappellari et al. (2002, ApJ, 578, 787)

    #    sigmapsf = [0.494, 1.44, 4.71, 13.4]      # In PC1 pixels
    #    normpsf = [0.294, 0.559, 0.0813, 0.0657]  # total(normpsf)=1

    parfile="sky.txt"
    outname=TNAM

    rmsname="none"
    psfname="none"

    consfile="none"

    T1 = "{}".format(imgname)
    T2 = outname + "-sky.fits"
    T3 = "{}".format(rmsname)

    if flagpos == False:
        xlo=1
        ylo=1

        (xhi,yhi)=GetAxis(imgname)

    plate=1

    fout1 = open(parfile, "w")

    PrintHeader(fout1, T1, T2, T3, psfname, 1, maskfile, consfile, xlo, xhi, ylo,
                yhi, convbox, convbox, mgzpt, plate, plate, "regular", 0, 0)

    index = 0

    index+=1



    PrintSky(fout1, index, sky, Z, fit)

    fout1.close()


    print("To compute sky run: galfit sky.txt")





####################################
#####################################
######################################

def ds9satbox (satfileout,output,satscale,satoffset):
    "Creates a file for ds9 which selects bad saturated regions"

    scaleflag=1
    offsetflag=1
    regfileflag=1
    magflag=1
    clasflag=1

    flagsat=4      ## flag value when object is saturated (or close to)
    maxflag=128    ## max value for flag
    check=0
    regflag = 0    ## flag for saturaded regions




    f_out = open(satfileout, "w")

    N,Alpha,Delta,X,Y,Mg,Kr,Fluxr,Isoa,Ai,E,Theta,Bkgd,Idx,Flg=np.genfromtxt(output,delimiter="",unpack=True)


    line="image \n"
    f_out.write(line)



    for idx, item in enumerate(N):


        bi=Ai[idx]*(1-E[idx])

        Theta[idx] = Theta[idx] * np.pi /180  #rads!!!


        Rkronx = satscale * 2 * Ai[idx] * Kr[idx]  + satoffset
        Rkrony = satscale * 2 * bi * Kr[idx]  + satoffset



        if Rkronx == 0:
            Rkronx = 1

        if Rkrony == 0:
            Rkrony = 1

        check=CheckFlag(Flg[idx],flagsat)  ## check if object has saturated regions

        if (check):

            line="box({0},{1},{2},{3},0) # color=red move=0 \n".format(X[idx],Y[idx],Rkronx,Rkrony)
            f_out.write(line)

            line2="point({0},{1}) # point=boxcircle font=\"times 10 bold\" text={{ {2} }} \n".format(X[idx],Y[idx],N[idx])
            f_out.write(line2)


    f_out.close()





def MakeMask(maskimage, catfile, scale, offset, regfile):
    "Create a mask image using ellipses for every Object of catfile. Now includes offset"
# k Check

    checkflag = 0
    flagsat = 4  # flag value when object is saturated (or close to)
    maxflag = 128  # max value for flag

    regflag = 0  # flag for saturaded regions

    n, alpha, delta, xx, yy, mg, kr, fluxrad, ia, ai, e, theta, bkgd, idx, flg, sxmin, sxmax, symin, symax, sxsmin, sxsmax, sysmin, sysmax = np.genfromtxt(
        catfile, delimiter="", unpack=True)

    n = n.astype(int)
    flg = flg.astype(int)

    print("Creating Masks for sky \n")

    Rkron = scale * ai * kr + offset

    mask = Rkron < 1
    if mask.any():
        Rkron[mask] = 1

    hdu = fits.open(maskimage)
    img = hdu[0].data

    for idx, val in enumerate(n):

        # check if object doesn't has saturaded regions
        checkflag = CheckFlag(flg[idx], flagsat)
        # check if object is inside of a saturaded box region indicated by
        # user in ds9
        regflag = CheckSatReg(xx[idx], yy[idx], regfile, Rkron[idx], theta[idx], e[idx])

        if (checkflag == False) and (regflag == False):

            print ("Creating ellipse mask for object {}  \n".format(n[idx]))
            img = MakeKron(img, n[idx], xx[idx], yy[idx], Rkron[idx], theta[idx], e[
                idx], sxsmin[idx], sxsmax[idx], sysmin[idx], sysmax[idx])

        elif(checkflag == True or regflag == True):

            print ("Skipping object {}, one or more pixels are saturated \n".format(n[idx]))

    hdu[0].data = img
    hdu.writeto(maskimage, overwrite=True)
    hdu.close()

    return True


def MakeKron(imagemat, idn, x, y, R, theta, ell, xmin, xmax, ymin, ymax):
    "This subroutine create a Kron ellipse within a box defined by: xmin, xmax, ymin, ymax"

# Check

    xmin = np.int(xmin)
    xmax = np.int(xmax)
    ymin = np.int(ymin)
    ymax = np.int(ymax)

    q = (1 - ell)
    bim = q * R

    theta = theta * np.pi / 180  # Rads!!!

    ypos, xpos = np.mgrid[ymin - 1:ymax, xmin - 1:xmax]

    dx = xpos - x
    dy = ypos - y

    landa = np.arctan2(dy, dx)

    mask = landa < 0
    if mask.any():
        landa[mask] = landa[mask] + 2 * np.pi

    landa = landa - theta

    angle = np.arctan2(np.sin(landa) / bim, np.cos(landa) / R)

    xell = x + R * np.cos(angle) * np.cos(theta) - bim * \
        np.sin(angle) * np.sin(theta)
    yell = y + R * np.cos(angle) * np.sin(theta) + bim * \
        np.sin(angle) * np.cos(theta)

    dell = np.sqrt((xell - x)**2 + (yell - y)**2)
    dist = np.sqrt(dx**2 + dy**2)

    mask = dist < dell
    imagemat[ypos[mask], xpos[mask]] = idn

    return imagemat


def MakeSatBox(maskimage, region, val, ncol, nrow):
    "Create a mask for saturated regions"
    "Regions must be in DS9 box regions format"

# k Check

#	fileflag=1

    hdu = fits.open(maskimage)
    img = hdu[0].data

    with open(region) as f_in:

        next(f_in)
#        next(f_in)
#        next(f_in)

        # All lines including the blank ones
        lines = (line.rstrip() for line in f_in)
        lines = (line.split('#', 1)[0] for line in lines)  # remove comments
        # remove lines containing only comments
        lines = (line.rstrip() for line in lines)
        lines = (line for line in lines if line)  # Non-blank lines

        for line in lines:

            (box, info) = line.split('(')

            if(box == "box"):

                (xpos, ypos, xlong, ylong, trash) = info.split(',')

                xpos = float(xpos)
                ypos = float(ypos)
                xlong = float(xlong)
                ylong = float(ylong)

                xlo = (xpos - xlong / 2)
                xhi = (xpos + xlong / 2)

                ylo = (ypos - ylong / 2)
                yhi = (ypos + ylong / 2)

                xlo = int(xlo)
                xhi = int(xhi)

                ylo = int(ylo)
                yhi = int(yhi)

                if (xlo < 1):

                    xlo = 1

                if (xhi > ncol):

                    xhi = ncol

                if (ylo < 1):

                    ylo = 1

                if (yhi > nrow):

                    yhi = nrow

                img[ylo - 1:yhi, xlo - 1:xhi] = val

    hdu[0].data = img
    hdu.writeto(maskimage, overwrite=True)
    hdu.close()

    return True




def MakeImage(newfits, sizex, sizey):
    "create a new blank Image"
# k Check
    if os.path.isfile(newfits):
        print("{} deleted; a new one is created \n".format(newfits))
    hdu = fits.PrimaryHDU()
    hdu.data = np.zeros((sizey, sizex))
    hdu.writeto(newfits, overwrite=True)

    return True



def CatArSort(SexCat,scale,SexArSort,NCol,NRow):
    # k Check

    # sort the sextractor
    # catalog by magnitude,
    # get sizes for objects
    # and write it in a new file

    # The sextractor catalog must contain the following parameters:
    #   1 NUMBER                 Running object number
    #   2 ALPHA_J2000            Right ascension of barycenter (J2000)                      [deg]
    #   3 DELTA_J2000            Declination of barycenter (J2000)                          [deg]
    #   4 X_IMAGE                Object position along x                                    [pixel]
    #   5 Y_IMAGE                Object position along y                                    [pixel]
    #   6 MAG_APER               Fixed aperture magnitude vector                            [mag]
    #   7 KRON_RADIUS            Kron apertures in units of A or B
    #   8 FLUX_RADIUS            Fraction-of-light radii                                    [pixel]
    #   9 ISOAREA_IMAGE          Isophotal area above Analysis threshold                    [pixel**2]
    #  10 A_IMAGE                Profile RMS along major axis                               [pixel]
    #  11 ELLIPTICITY            1 - B_IMAGE/A_IMAGE
    #  12 THETA_IMAGE            Position angle (CCW/x)                                     [deg]
    #  13 BACKGROUND             Background at centroid position                            [count]
    #  14 CLASS_STAR             S/G classifier output
    #  15 FLAGS                  Extraction flags


    print("Sorting and getting sizes for objects \n")

    n, alpha, delta, xx, yy, mg, kr, fluxrad, ia, ai, e, theta, bkgd, idx, flg = np.genfromtxt(
        SexCat, delimiter="", unpack=True)

    n = n.astype(int)
    flg = flg.astype(int)

#    ai = ai.astype(float)
#    kr = kr.astype(float)

#    scale = scale.astype(float)


    Rkron = scale * ai * kr

    Rwsky = scale * ai * kr + 10  + 20


#   considering to use only  KronScale instead of SkyScale
#    Rwsky = parvar.KronScale * ai * kr + parvar.Offset + parvar.SkyWidth

    Bim = (1 - e) * Rkron

    Area = np.pi * Rkron * Bim *(-1)

    (sxmin, sxmax, symin, symax) = GetSize(xx, yy, Rkron, theta, e, NCol, NRow)

    (sxsmin, sxsmax, sysmin, sysmax) = GetSize(xx, yy, Rwsky, theta, e, NCol,NRow)

#    print(sxmin.size, sxmax.size, symin, symax)

    f_out = open(SexArSort, "w")

    index = Area.argsort()
    for i in index:

        line = "{} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {} {}\n".format(n[i], alpha[i], delta[i], xx[i], yy[i], mg[i], kr[i], fluxrad[i], ia[i], ai[i], e[i], theta[i], bkgd[i], idx[i], flg[i], np.int(
            np.round(sxmin[i])), np.int(np.round(sxmax[i])), np.int(np.round(symin[i])), np.int(np.round(symax[i])), np.int(np.round(sxsmin[i])), np.int(np.round(sxsmax[i])), np.int(np.round(sysmin[i])), np.int(np.round(sysmax[i])))

        f_out.write(line)

    f_out.close()

    return len(n)


def GetSize(x, y, R, theta, ell, ncol, nrow):
    "this subroutine get the maximun"
    "and minimim pixels for Kron and sky ellipse"
    # k Check
    q = (1 - ell)
    bim = q * R

    theta = theta * (np.pi / 180)  # rads!!

# getting size

    xmin = x - np.sqrt((R**2) * (np.cos(theta))**2 +
                       (bim**2) * (np.sin(theta))**2)

    xmax = x + np.sqrt((R**2) * (np.cos(theta))**2 +
                       (bim**2) * (np.sin(theta))**2)

    ymin = y - np.sqrt((R**2) * (np.sin(theta))**2 +
                       (bim**2) * (np.cos(theta))**2)

    ymax = y + np.sqrt((R**2) * (np.sin(theta))**2 +
                       (bim**2) * (np.cos(theta))**2)

    mask = xmin < 1
    if mask.any():
        xmin[mask] = 1

    mask = xmax > ncol

    if mask.any():
        xmax[mask] = ncol

    mask = ymin < 1
    if mask.any():
        ymin[mask] = 1

    mask = ymax > nrow
    if mask.any():
        ymax[mask] = nrow

    return (xmin, xmax, ymin, ymax)

#def CheckFlag(val,check,max):
 #  "Check for flag contained in $val, returns 1 if found "

  # flag = False
  # mod = 1

   #while (mod != 0):


    #   res = int(val/max)

    #3  if (max == check and res == 1 ):

    #       flag=True


     #  mod = val % max

      # val = mod
       #max = max/2


#   return flag


def CheckSatReg(x,y,filein,R,theta,ell):
   "Check if object is inside of saturated region. returns True if at least one pixel is inside"
## check if object is inside of
## saturaded region as indicated by ds9 box region
## returns 1 if object center is in saturaded region


   q = (1 - ell)

   bim = q * R

   theta = theta * np.pi /180  ## Rads!!!

   flag = False
   fileflag =1



   with open(filein) as f_in:

       lines = (line.rstrip() for line in f_in) # All lines including the blank ones
       lines = (line.split('#', 1)[0] for line in lines) # remove comments
       lines = (line.rstrip() for line in lines)   # remove lines containing only comments
       lines = (line for line in lines if line) # Non-blank lines

       for line in lines:


           if (line != "image"):

               (box,info)=line.split('(')

               if(box == "box"):

                   (xpos,ypos,xlong,ylong,trash)=info.split(',')

                   xpos=float(xpos)
                   ypos=float(ypos)
                   xlong=float(xlong)
                   ylong=float(ylong)


                   xlo = xpos - xlong/2
                   xhi = xpos + xlong/2

                   ylo = ypos - ylong/2
                   yhi = ypos + ylong/2

                   dx = xpos - x
                   dy = ypos - y

                   landa=np.arctan2( dy,dx )

                   if landa < 0:
                       landa=landa + 2 * np.pi


                   landa = landa - theta

                   angle = np.arctan2(np.sin(landa)/bim, np.cos(landa)/R)

                   xell =  x + R * np.cos(angle)* np.cos(theta)  - bim * np.sin(angle) * np.sin(theta)
                   yell =  y + R * np.cos(angle)* np.sin(theta)  + bim * np.sin(angle) * np.cos(theta)


                   if ( (xell > xlo and xell < xhi) and (yell > ylo and yell < yhi)  ):

                       flag=True
                       break


   return flag













#################333333333333333333333333333333333
####################################################
####################################################
#####################################################





### GALFIT functions

def PrintHeader(hdl, A, B, C, D, E, F, G, xlo, xhi, ylo, yhi, convx, convy, J, platedx, platedy, O, P, S):
    "print GALFIT header in a file"

    # k Check
    # print to filehandle
    # the header for GALFIT

    lineZ = "==================================================================================================\n"
    lineX = "# IMAGE PARAMETERS \n"
    lineA = "A) {}                                   # Input Data image (FITS file)                            \n".format(A)
    lineB = "B) {}                                   # Output data image block                                 \n".format(B)
    lineC = "C) {}                                   # Sigma image name (made from data if blank or \"none\")  \n".format(C)
    lineD = "D) {}                                   # Input PSF image and (optional) diffusion kernel         \n".format(D)
    lineE = "E) {}                                   # PSF fine sampling factor relative to data               \n".format(E)
    lineF = "F) {}                                   # Bad pixel mask (FITS image or ASCII coord list)         \n".format(F)
    lineG = "G) {}                                   # File with parameter constraints (ASCII file)            \n".format(G)
    lineH = "H) {} {} {} {}                          # Image region to fit (xmin xmax ymin ymax)               \n".format(xlo, xhi, ylo, yhi)
    lineI = "I) {} {}                                # Size of the convolution box (x y)                       \n".format(convx, convy)
    lineJ = "J) {}                                   # Magnitude photometric zeropoint                         \n".format(J)
    lineK = "K) {} {}                                # Plate scale (dx dy). \[arcsec per pixel\]               \n".format(platedx, platedy)
    lineO = "O) {}                                   # Display type (regular, curses, both)                    \n".format(O)
    lineP = "P) {}                                   # Choose 0=optimize, 1=model, 2=imgblock, 3=subcomps      \n".format(P)
    lineS = "S) {}                                   # Modify/create objects interactively?                    \n".format(S)
    lineY = " \n"

    line0 = "# INITIAL FITTING PARAMETERS                                                     \n"
    line1 = "# \n"
    line2 = "#   For object type, allowed functions are:                                      \n"
    line3 = "#       nuker, sersic, expdisk, devauc, king, psf, gaussian, moffat,             \n"
    line4 = "#       ferrer, powsersic, sky, and isophote.                                    \n"
    line5 = "# \n"
    line6 = "#  Hidden parameters will only appear when they're specified:                    \n"
    line7 = "#      C0 (diskyness/boxyness),                                                  \n"
    line8 = "#      Fn (n=integer, Azimuthal Fourier Modes),                                  \n"
    line9 = "#      R0-R10 (PA rotation, for creating spiral structures).                     \n"
    line10 = "# \n"

    line11 = "# column 1:  Parameter number                                                               \n"
    line12 = "# column 2:                                                                                 \n"
    line13 = "#          -- Parameter 0:    the allowed functions are: sersic, nuker, expdisk             \n"
    line14 = "#                             edgedisk, devauc, king, moffat, gaussian, ferrer, psf, sky    \n"
    line15 = "#          -- Parameter 1-10: value of the initial parameters                               \n"
    line16 = "#          -- Parameter C0:   For diskiness/boxiness                                        \n"
    line17 = "#                             <0 = disky                                                    \n"
    line18 = "#                             >0 = boxy                                                     \n"
    line19 = "#          -- Parameter Z:    Outputting image options, the options are:                    \n"
    line20 = "#                             0 = normal, i.e. subtract final model from the data to create \n"
    line21 = "#                             the residual image                                            \n"
    line22 = "#                             1 = Leave in the model -- do not subtract from the data       \n"
    line23 = "#                                                                                           \n"
    line24 = "# column 3: allow parameter to vary (yes = 1, no = 0)                                       \n"
    line25 = "# column 4: comment                                                                         \n"
    line26 = " \n"

    line27 = "==================================================================================================\n"

    hdl.write(lineZ)
    hdl.write(lineX)
    hdl.write(lineA)
    hdl.write(lineB)
    hdl.write(lineC)
    hdl.write(lineD)
    hdl.write(lineE)
    hdl.write(lineF)
    hdl.write(lineG)
    hdl.write(lineH)
    hdl.write(lineI)
    hdl.write(lineJ)
    hdl.write(lineK)
    hdl.write(lineO)
    hdl.write(lineP)
    hdl.write(lineS)
    hdl.write(lineY)

    hdl.write(line0)
    hdl.write(line1)
    hdl.write(line2)
    hdl.write(line3)
    hdl.write(line4)
    hdl.write(line5)
    hdl.write(line6)
    hdl.write(line7)
    hdl.write(line8)
    hdl.write(line9)
    hdl.write(line10)

    hdl.write(line11)
    hdl.write(line12)
    hdl.write(line13)
    hdl.write(line14)
    hdl.write(line15)
    hdl.write(line16)
    hdl.write(line17)
    hdl.write(line18)
    hdl.write(line19)
    hdl.write(line20)
    hdl.write(line21)
    hdl.write(line22)
    hdl.write(line23)
    hdl.write(line24)
    hdl.write(line25)
    hdl.write(line26)
    hdl.write(line27)

    return True


def PrintSky(hdl, ncomp, sky, Z, fit):
    "Print GALFIT sky function to filehandle"

    # k Check

    line00 = "# Object number: {}                                                             \n".format(ncomp)
    line01 = " 0)      sky            #    Object type                                        \n"
    line02 = " 1) {}         {}       # sky background        [ADU counts]                    \n".format(sky, fit)
    line03 = " 2) 0.000      0        # dsky/dx (sky gradient in x)                           \n"
    line04 = " 3) 0.000      0        # dsky/dy (sky gradient in y)                           \n"
    line05 = " Z) {}                  # Skip this model in output image?  (yes=1, no=0)       \n".format(Z)
    line06 = "\n"
    line07 = "================================================================================\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)

    return True


def PrintSersic(hdl, ncomp, xpos, ypos, magser, reser, nser, axratser, angleser, Z, fit):
    "print GALFIT Sersic function to filehandle"
    # k Check

    # print to filehandle
    # a sersic function given
    # by the parameters

    line00 = "# Object number: {}                                                             \n".format(
            ncomp)
    line01 = " 0)     sersic               #  Object type                                     \n"
    line02 = " 1) {:.2f}  {:.2f}  {}  {}            #  position x, y     [pixel]                       \n".format(
        xpos, ypos, fit, fit)
    line03 = " 3) {:.2f}       {}              #  total magnitude                                 \n".format(
        magser, fit)
    line04 = " 4) {:.2f}       {}              #  R_e         [Pixels]                            \n".format(
            reser, fit)
    line05 = " 5) {}       {}              #  Sersic exponent (deVauc=4, expdisk=1)           \n".format(
        nser, fit)
    line06 = " 9) {:.2f}       {}              #  axis ratio (b/a)                                \n".format(
        axratser, fit)
    line07 = "10) {:.2f}       {}              #  position angle (PA)  [Degrees: Up=0, Left=90]   \n".format(
        angleser, fit)
    line08 = " Z) {}                       #  Skip this model in output image?  (yes=1, no=0) \n".format(
        Z)
    line09 = "\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)
    hdl.write(line08)
    hdl.write(line09)

    return True

def PrintGauss(hdl, ncomp, xpos, ypos, magass, fwhm, axratgass, anglegass, Z, fit):
    "print GALFIT GAUSS function to filehandle"

    # print to filehandle
    # a sersic function given
    # by the parameters

    line00 = "# Object number: {}                                                             \n".format(
            ncomp)
    line01 = " 0)     gaussian               #  Object type                                     \n"
    line02 = " 1) {:.2f}  {:.2f}  {}  {}            #  position x, y     [pixel]                       \n".format(
        xpos, ypos, fit, fit)
    line03 = " 3) {:.2f}       {}              #  total magnitude                                 \n".format(
        magass, fit)
    line04 = " 4) {:.2f}       {}              #  FWHM         [Pixels]                            \n".format(
            fwhm, fit)
    line05 = " 9) {:.2f}       {}              #  axis ratio (b/a)                                \n".format(
        axratgass, fit)
    line06 = "10) {:.2f}       {}              #  position angle (PA)  [Degrees: Up=0, Left=90]   \n".format(
        anglegass, fit)
    line07 = " Z) {}                       #  Skip this model in output image?  (yes=1, no=0) \n".format(
        Z)
    line08 = "\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)
    hdl.write(line08)

    return True

def PrintExp(hdl, ncomp, xpos, ypos, magexp, rsexp, axratexp, angleexp, Z, fit):
    "print GALFIT exponential function to filehandle"

    # k Check

    # print to filehandle
    # a exponential function given
    # by the parameters

    line00 = "# Object number: $ncomp                                                        \n"
    line01 = " 0)     expdisk              # Object type                                     \n"
    line02 = " 1) {:.2f}  {:.2f}  {}  {}           # position x, y     [pixel]                       \n".format(
        xpos, ypos, fit, fit)
    line03 = " 3) {:.2f}        {}             # total magnitude                                 \n".format(
        magexp, fit)
    line04 = " 4) {:.2f}        {}             #      Rs  [Pixels]                               \n".format(
        rsexp, fit)
    line05 = " 9) {:.2f}        {}             # axis ratio (b/a)                                \n".format(
        axratexp, fit)
    line06 = "10) {:.2f}        {}             # position angle (PA)  [Degrees: Up=0, Left=90]   \n".format(
        angleexp, fit)
    line07 = " Z) {}                       # Skip this model in output image?  (yes=1, no=0) \n".format(
        Z)
    line08 = "\n"

    hdl.write(line00)
    hdl.write(line01)
    hdl.write(line02)
    hdl.write(line03)
    hdl.write(line04)
    hdl.write(line05)
    hdl.write(line06)
    hdl.write(line07)
    hdl.write(line08)

    return True


def GetFits(Image, Imageout, xlo, xhi, ylo, yhi):
    "Get a piece from the image"
# k Check


    if os.path.isfile(Imageout):
        print("{} deleted; a new one is created \n".format(Imageout))
        runcmd = "rm {}".format(Imageout)
        errrm = sp.run([runcmd], shell=True, stdout=sp.PIPE,
                        stderr=sp.PIPE, universal_newlines=True)


    hdu = fits.open(Image)
    dat = hdu[0].data[ylo - 1:yhi, xlo - 1:xhi]
    hdu[0].data = dat
    hdu.writeto(Imageout, overwrite=True)
    hdu.close()

def GetExpTime(Image):
    # k Check
    "Get exposition time from the image"

    hdu = fits.open(Image)
    exptime = hdu[0].header["EXPTIME"]
    hdu.close()
    return exptime


def CheckFlag(val, check):
    "Check for flag contained in val, returns True if found "

    flag = False
    mod = 1
    maxx=128


    while (mod != 0):

        res = int(val / maxx)

        if (maxx == check and res == 1):

            flag = True

        mod = val % maxx

        val = mod
        maxx = maxx / 2

    return flag


#############################################################################
######################### End of program  ###################################
#     ______________________________________________________________________
#    /___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/___/_/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/|
#   |___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|__/|
#   |_|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|___|/
##############################################################################
if __name__ == '__main__':
    mainGalfitSky()




import galfitools 


def printWelcome():


    print("GALFITools: a library for GALFIT")

    print("Version:",galfitools.__version__)

    url = "https://github.com/canorve/GALFITools"
    
    print("webpage: "+url+"\n")


